"""
.. module:: theories.classy

:Synopsis: Managing the CLASS cosmological code
:Author: Jesus Torrado
         (import and ``get_Cl`` based on MontePython's CLASS wrapper Benjamin Audren)

.. |br| raw:: html

   <br />

This module imports and manages the CLASS cosmological code.

.. note::

   **If you use this cosmological code, please cite it as:**
   |br|
   D. Blas, J. Lesgourgues, T. Tram,
   *The Cosmic Linear Anisotropy Solving System (CLASS). Part II: Approximation schemes*
   (`arXiv:1104.2933 <https://arxiv.org/abs/1104.2933>`_)

.. note::

   CLASS is renamed ``classy`` for most purposes within cobaya, due to CLASS's name being
   a python keyword.

Usage
-----

If you are using a likelihood that requires some observable from CLASS, simply add
``classy`` to the theory block.

You can specify any parameter that CLASS understands in the ``params`` block:

.. code-block:: yaml

   theory:
     classy:
       extra_args:
         [any param that CLASS understands]

   params:
       [any param that CLASS understands, fixed, sampled or derived]

If you want to use your own version of CLASS, you need to specify its location with a
``path`` option inside the ``classy`` block. If you do not specify a ``path``,
CLASS will be loaded from the automatic-install ``packages_path`` folder, if specified, or
otherwise imported as a globally-installed Python package. If you want to force that
the global ``classy`` installation is used, pass ``path='global'``. Cobaya will print at
initialisation where CLASS was actually loaded from.


.. _classy_access:

Access to CLASS computation products
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

You can retrieve CLASS computation products within likelihoods (or other pipeline
components in general) or manually from a :class:`~model.Model` as long as you have added
them as requisites; see :doc:`cosmo_external_likelihood` or
:doc:`cosmo_external_likelihood_class` for the likelihood case, and :doc:`cosmo_model` for
the manual case.

The products that you can request and later retrieve are listed in
:func:`~theories.cosmo.BoltzmannBase.must_provide`.

If you would like to access a CLASS result that is not accessible that way, you can access
directly the return value of the `python CLASS interface
<https://github.com/lesgourg/class_public/blob/master/python/classy.pyx>`_
``get_background()``, ``get_thermodynamics()``, ``get_primordial()``,
``get_perturbations()`` and  ``get_sources()``. To do that add to the requisites
``{get_CLASS_[...]: None}`` respectively, and retrieve it with
``provider.get_CLASS_[...]``.

In general, the use of these methods for direct access to CLASS results should be avoided
in public code, since it breaks compatibility with other Boltzmann codes at the likelihood
interface level. If you need a quantity for a public code that is not generally interfaced
in :func:`~theories.cosmo.BoltzmannBase.must_provide`, let us know if you think it makes
sense to add it.


.. _classy_modify:

Modifying CLASS
^^^^^^^^^^^^^^^

If you modify CLASS and add new variables, make sure that the variables you create are
exposed in the Python interface
(`instructions here <https://github.com/lesgourg/class_public/wiki/Python-wrapper>`__).
If you follow those instructions you do not need to make any additional modification in
Cobaya.

If your modification involves new computed quantities, add the new quantities to the
return value of some of the direct-access methods listed in :ref:`classy_access`.

You can use the :doc:`model wrapper <cosmo_model>` to test your modification by
evaluating observables or getting derived quantities at known points in the parameter
space (set ``debug: True`` to get more detailed information of what exactly is passed to
CLASS).

In your CLASS modification, remember that you can raise a ``CosmoComputationError``
whenever the computation of any observable would fail, but you do not expect that
observable to be compatible with the data (e.g. at the fringes of the parameter
space). Whenever such an error is raised during sampling, the likelihood is assumed to be
zero, and the run is not interrupted.

.. note::

   If your modified CLASS has a lower version number than the minimum required by Cobaya,
   you will get an error at initialisation. You may still be able to use it by setting the
   option ``ignore_obsolete: True`` in the ``classy`` block (though you would be doing
   that at your own risk; ideally you should translate your modification to a newer CLASS
   version, in case there have been important fixes since the release of your baseline
   version).


Installation
------------

   .. _classy_install_warn:

.. warning::

   If the installation folder of CLASS is moved, due to CLASS hard-coding some folders,
   CLASS needs to be recompiled, either manually or by deleting the CLASS installation and
   repeating the ``cobaya-install`` command in the renamed *packages* folder.

   If you do not recompile CLASS, it causes a memory leak (`thanks to Stefan Heimersheim
   <https://github.com/CobayaSampler/cobaya/issues/10>`_).

Automatic installation
^^^^^^^^^^^^^^^^^^^^^^

If you do not plan to modify CLASS, the easiest way to install it is using the
:doc:`automatic installation script <installation_cosmo>`. Just make sure that
``theory: classy:`` appears in one of the files passed as arguments to the installation
script.

Manual installation (or using your own version)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you are planning to modify CLASS or use an already modified version,
you should not use the automatic installation script. Use the method below instead.

CLASS's python interface utilizes the ``cython`` compiler. If typing ``cython`` in the
shell produces an error, install it with ``python -m pip install cython``.

.. note::
   The fast way, assuming you are installing all your cosmological codes under
   ``/path/to/cosmo/``:

   .. code:: bash

      $ cd /path/to/cosmo/
      $ git clone https://github.com/lesgourg/class_public.git
      $ mv class_public CLASS
      $ cd CLASS
      $ make

   If the **second** line produces an error (because you don't have ``git`` installed),
   try

   .. code:: bash

      $ cd /path/to/cosmo/
      $ wget https://github.com/lesgourg/class_public/archive/master.zip
      $ unzip master.zip
      $ rm master.zip
      $ mv class_public-master CLASS
      $ cd CLASS
      $ make

If the instructions above failed, follow those in the
`official CLASS web page <http://class-code.net/>`_ to get CLASS compiled with the Python
interface ready.
"""

# Global
import sys
import os
import numpy as np
from copy import deepcopy
from typing import NamedTuple, Sequence, Union, Optional, Callable, Any

# Local
from cobaya.theories.cosmo import BoltzmannBase
from cobaya.log import LoggedError, get_logger
from cobaya.install import download_github_release, pip_install, check_gcc_version
from cobaya.component import ComponentNotInstalledError, load_external_module
from cobaya.tools import Pool1D, Pool2D, PoolND, combine_1d, get_compiled_import_path, \
    VersionCheckError


# Result collector
# NB: cannot use kwargs for the args, because the CLASS Python interface
#     is C-based, so args without default values are not named.
class Collector(NamedTuple):
    method: str
    args: Sequence = []
    args_names: Sequence = []
    kwargs: dict = {}
    arg_array: Union[int, Sequence, None] = None
    z_pool: Optional[PoolND] = None
    post: Optional[Callable] = None


# default non linear code -- same as CAMB
non_linear_default_code = "hmcode"
non_linear_null_value = "none"


class classy(BoltzmannBase):
    r"""
    CLASS cosmological Boltzmann code \cite{Blas:2011rf}.
    """

    # Name of the Class repo/folder and version to download
    _classy_repo_name = "lesgourg/class_public"
    _min_classy_version = "v3.2.0"
    _classy_min_gcc_version = "6.4"  # Lower ones are possible atm, but leak memory!
    _classy_repo_version = os.environ.get('CLASSY_REPO_VERSION', _min_classy_version)

    classy_module: Any
    ignore_obsolete: bool

    def initialize(self):
        """Importing CLASS from the correct path, if given, and if not, globally."""
        try:
            install_path = (lambda p: self.get_path(p) if p else None)(self.packages_path)
            min_version = None if self.ignore_obsolete else self._classy_repo_version
            self.classy_module = load_external_module(
                "classy", path=self.path, install_path=install_path,
                min_version=min_version, get_import_path=self.get_import_path,
                logger=self.log, not_installed_level="debug")
        except VersionCheckError as excpt:
            raise VersionCheckError(
                str(excpt) + " If you are using CLASS unmodified, upgrade with"
                "`cobaya-install classy --upgrade`. If you are using a modified CLASS, "
                "set the option `ignore_obsolete: True` for CLASS.")
        except ComponentNotInstalledError as excpt:
            raise ComponentNotInstalledError(
                self.log, (f"Could not find CLASS: {excpt}. "
                           "To install it, run `cobaya-install classy`"))
        self.classy = self.classy_module.Class()
        super().initialize()
        # Add general CLASS stuff
        self.extra_args["output"] = self.extra_args.get("output", "")
        if "sBBN file" in self.extra_args:
            self.extra_args["sBBN file"] = (
                self.extra_args["sBBN file"].format(classy=self.path))
        # Normalize `non_linear` vs `non linear`: prefer underscore
        # Keep this convention throughout the rest of this module!
        if "non linear" in self.extra_args:
            if "non_linear" in self.extra_args:
                raise LoggedError(
                    self.log, ("In `extra_args`, only one of `non_linear` or `non linear`"
                               " should be defined."))
            self.extra_args["non_linear"] = self.extra_args.pop("non linear")
        # Normalize non_linear None|False --> "none"
        # Use default one if not specified
        if self.extra_args.get("non_linear", "dummy_string") in (None, False):
            self.extra_args["non_linear"] = non_linear_null_value
        elif ("non_linear" not in self.extra_args or
              self.extra_args["non_linear"] is True):
            self.extra_args["non_linear"] = non_linear_default_code
        # Derived parameters that may not have been requested, but will be necessary later
        self.derived_extra = []

    def set_cl_reqs(self, reqs):
        """
        Sets some common settings for both lensend and unlensed Cl's.
        """
        if any(("t" in cl.lower()) for cl in reqs):
            self.extra_args["output"] += " tCl"
        if any((("e" in cl.lower()) or ("b" in cl.lower())) for cl in reqs):
            self.extra_args["output"] += " pCl"
        # For l_max_scalars, remember previous entries.
        self.extra_args["l_max_scalars"] = \
            max(self.extra_args.get("l_max_scalars", 0), max(reqs.values()))
        if 'T_cmb' not in self.derived_extra:
            self.derived_extra += ['T_cmb']

    def must_provide(self, **requirements):
        # Computed quantities required by the likelihood
        super().must_provide(**requirements)
        for k, v in self._must_provide.items():
            # Products and other computations
            if k == "Cl":
                self.set_cl_reqs(v)
                # For modern experiments, always lensed Cl's!
                self.extra_args["output"] += " lCl"
                self.extra_args["lensing"] = "yes"
                self.collectors[k] = Collector(
                    method="lensed_cl", kwargs={"lmax": self.extra_args["l_max_scalars"]})
            elif k == "unlensed_Cl":
                self.set_cl_reqs(v)
                self.collectors[k] = Collector(
                    method="raw_cl", kwargs={"lmax": self.extra_args["l_max_scalars"]})
            elif k == "Hubble":
                self.set_collector_with_z_pool(
                    k, v["z"], "Hubble", args_names=["z"], arg_array=0)
            elif k in ["Omega_b", "Omega_cdm", "Omega_nu_massive"]:
                func_name = {"Omega_b": "Om_b", "Omega_cdm": "Om_cdm",
                             "Omega_nu_massive": "Om_ncdm"}[k]
                self.set_collector_with_z_pool(
                    k, v["z"], func_name, args_names=["z"], arg_array=0)
            elif k == "angular_diameter_distance":
                self.set_collector_with_z_pool(
                    k, v["z"], "angular_distance", args_names=["z"], arg_array=0)
            elif k == "comoving_radial_distance":
                self.set_collector_with_z_pool(k, v["z"], "z_of_r", args_names=["z"],
                                               # returns r and dzdr!
                                               post=(lambda r, dzdr: r))
            elif k == "angular_diameter_distance_2":
                self.set_collector_with_z_pool(
                    k, v["z_pairs"], "angular_distance_from_to",
                    args_names=["z1", "z2"], arg_array=[0, 1], d=2)
            elif isinstance(k, tuple) and k[0] == "Pk_grid":
                self.extra_args["output"] += " mPk"
                v = deepcopy(v)
                self.add_P_k_max(v.pop("k_max"), units="1/Mpc")
                # NB: Actually, only the max z is used, and the actual sampling in z
                # for computing P(k,z) is controlled by `perturb_sampling_stepsize`
                # (default: 0.1). But let's leave it like this in case this changes
                # in the future.
                self.add_z_for_matter_power(v.pop("z"))
                if v["nonlinear"]:
                    if "non_linear" not in self.extra_args:
                        # this is redundant with initialisation, but just in case
                        self.extra_args["non_linear"] = non_linear_default_code
                    elif self.extra_args["non_linear"] == non_linear_null_value:
                        raise LoggedError(
                            self.log, ("Non-linear Pk requested, but `non_linear: "
                                       f"{non_linear_null_value}` imposed in "
                                       "`extra_args`"))
                pair = k[2:]
                if pair == ("delta_tot", "delta_tot"):
                    v["only_clustering_species"] = False
                    self.collectors[k] = Collector(
                        method="get_pk_and_k_and_z",
                        kwargs=v,
                        post=(lambda P, kk, z: (kk, z, np.array(P).T)))
                elif pair == ("delta_nonu", "delta_nonu"):
                    v["only_clustering_species"] = True
                    self.collectors[k] = Collector(
                        method="get_pk_and_k_and_z", kwargs=v,
                        post=(lambda P, kk, z: (kk, z, np.array(P).T)))
                elif pair == ("Weyl", "Weyl"):
                    self.extra_args["output"] += " mTk"
                    self.collectors[k] = Collector(
                        method="get_Weyl_pk_and_k_and_z", kwargs=v,
                        post=(lambda P, kk, z: (kk, z, np.array(P).T)))
                else:
                    raise LoggedError(self.log, "NotImplemented in CLASS: %r", pair)
            elif k == "sigma8_z":
                self.add_z_for_matter_power(v["z"])
                self.set_collector_with_z_pool(
                    k, v["z"], "sigma", args=[8], args_names=["R", "z"],
                    kwargs={"h_units": True}, arg_array=1)
            elif k == "fsigma8":
                self.add_z_for_matter_power(v["z"])
                z_step = 0.1  # left to CLASS default; increasing does not appear to help
                self.set_collector_with_z_pool(
                    k, v["z"], "effective_f_sigma8", args=[z_step],
                    args_names=["z", "z_step"], arg_array=0)
            elif isinstance(k, tuple) and k[0] == "sigma_R":
                self.extra_args["output"] += " mPk"
                self.add_P_k_max(v.pop("k_max"), units="1/Mpc")
                # NB: See note about redshifts in Pk_grid
                self.add_z_for_matter_power(v["z"])
                pair = k[1:]
                try:
                    method = {("delta_tot", "delta_tot"): "sigma",
                              ("delta_nonu", "delta_nonu"): "sigma_cb"}[pair]
                except KeyError:
                    raise LoggedError(self.log, f"sigma(R,z) not implemented for {pair}")
                self.collectors[k] = Collector(
                    method=method, kwargs={"h_units": False}, args=[v["R"], v["z"]],
                    args_names=["R", "z"], arg_array=[[0], [1]],
                    post=(lambda R, z, sigma: (z, R, sigma.T)))
            elif k in [f"CLASS_{q}" for q in ["background", "thermodynamics",
                                              "primordial", "perturbations", "sources"]]:
                # Get direct CLASS results
                self.collectors[k] = Collector(method=f"get_{k.lower()[len('CLASS_'):]}")
            elif v is None:
                k_translated = self.translate_param(k)
                if k_translated not in self.derived_extra:
                    self.derived_extra += [k_translated]
            else:
                raise LoggedError(self.log, "Requested product not known: %r", {k: v})
        # Derived parameters (if some need some additional computations)
        if any(("sigma8" in s) for s in set(self.output_params).union(requirements)):
            self.extra_args["output"] += " mPk"
            self.add_P_k_max(1, units="1/Mpc")
        # Adding tensor modes if requested
        if self.extra_args.get("r") or "r" in self.input_params:
            self.extra_args["modes"] = "s,t"
        # If B spectrum with l>50, or lensing, recommend using a non-linear code
        cls = self._must_provide.get("Cl", {})
        has_BB_l_gt_50 = (any(("b" in cl.lower()) for cl in cls) and
                          max(cls[cl] for cl in cls if "b" in cl.lower()) > 50)
        has_lensing = any(("p" in cl.lower()) for cl in cls)
        if (has_BB_l_gt_50 or has_lensing) and \
           self.extra_args.get("non_linear") == non_linear_null_value:
            self.log.warning("Requesting BB for ell>50 or lensing Cl's: "
                             "using a non-linear code is recommended (and you are not "
                             "using any). To activate it, set "
                             "'non_linear: halofit|hmcode|...' in classy's 'extra_args'.")
        # Cleanup of products string
        self.extra_args["output"] = " ".join(set(self.extra_args["output"].split()))
        self.check_no_repeated_input_extra()

    def add_z_for_matter_power(self, z):
        if getattr(self, "z_for_matter_power", None) is None:
            self.z_for_matter_power = np.empty(0)
        self.z_for_matter_power = np.flip(combine_1d(z, self.z_for_matter_power))
        self.extra_args["z_pk"] = " ".join(["%g" % zi for zi in self.z_for_matter_power])

    def set_collector_with_z_pool(self, k, zs, method, args=(), args_names=(),
                                  kwargs=None, arg_array=None, post=None, d=1):
        """
        Creates a collector for a z-dependent quantity, keeping track of the pool of z's.

        If ``z`` is an arg, i.e. it is in ``args_names``, then omit it in the ``args``,
        e.g. ``args_names=["a", "z", "b"]`` should be passed together with
        ``args=[a_value, b_value]``.
        """
        if k in self.collectors:
            z_pool = self.collectors[k].z_pool
            z_pool.update(zs)
        else:
            Pool = {1: Pool1D, 2: Pool2D}[d]
            z_pool = Pool(zs)
        # Insert z as arg or kwarg
        kwargs = kwargs or {}
        if d == 1 and "z" in kwargs:
            kwargs = deepcopy(kwargs)
            kwargs["z"] = z_pool.values
        elif d == 1 and "z" in args_names:
            args = deepcopy(args)
            i_z = args_names.index("z")
            args = list(args[:i_z]) + [z_pool.values] + list(args[i_z:])
        elif d == 2 and "z1" in args_names and "z2" in args_names:
            # z1 assumed appearing before z2!
            args = deepcopy(args)
            i_z1 = args_names.index("z1")
            i_z2 = args_names.index("z2")
            args = (list(args[:i_z1]) + [z_pool.values[:, 0]] + list(args[i_z1:i_z2]) +
                    [z_pool.values[:, 1]] + list(args[i_z2:]))
        else:
            raise LoggedError(
                self.log,
                f"I do not know how to insert the redshift for collector method {method} "
                f"of requisite {k}")
        self.collectors[k] = Collector(
            method=method, z_pool=z_pool, args=args, args_names=args_names, kwargs=kwargs,
            arg_array=arg_array, post=post)

    def add_P_k_max(self, k_max, units):
        r"""
        Unifies treatment of :math:`k_\mathrm{max}` for matter power spectrum:
        ``P_k_max_[1|h]/Mpc``.

        Make ``units="1/Mpc"|"h/Mpc"``.
        """
        # Fiducial h conversion (high, though it may slow the computations)
        h_fid = 1
        if units == "h/Mpc":
            k_max *= h_fid
        # Take into account possible manual set of P_k_max_***h/Mpc*** through extra_args
        k_max_old = self.extra_args.pop(
            "P_k_max_1/Mpc", h_fid * self.extra_args.pop("P_k_max_h/Mpc", 0))
        self.extra_args["P_k_max_1/Mpc"] = max(k_max, k_max_old)

    def set(self, params_values_dict):
        # If no output requested, remove arguments that produce an error
        # (e.g. complaints if halofit requested but no Cl's computed.) ?????
        # Needed for facilitating post-processing
        if not self.extra_args["output"]:
            for k in ["non_linear"]:
                self.extra_args.pop(k, None)
        # Prepare parameters to be passed: this-iteration + extra
        args = {self.translate_param(p): v for p, v in params_values_dict.items()}
        args.update(self.extra_args)
        # Generate and save
        self.log.debug("Setting parameters: %r", args)
        self.classy.set(**args)

    def calculate(self, state, want_derived=True, **params_values_dict):
        # Set parameters
        self.set(params_values_dict)
        # Compute!
        try:
            self.classy.compute()
        # "Valid" failure of CLASS: parameters too extreme -> log and report
        except self.classy_module.CosmoComputationError as e:
            if self.stop_at_error:
                self.log.error(
                    "Computation error (see traceback below)! "
                    "Parameters sent to CLASS: %r and %r.\n"
                    "To ignore this kind of error, make 'stop_at_error: False'.",
                    state["params"], dict(self.extra_args))
                raise
            else:
                self.log.debug("Computation of cosmological products failed. "
                               "Assigning 0 likelihood and going on. "
                               "The output of the CLASS error was %s" % e)
            return False
        # CLASS not correctly initialized, or input parameters not correct
        except self.classy_module.CosmoSevereError:
            self.log.error("Serious error setting parameters or computing results. "
                           "The parameters passed were %r and %r. To see the original "
                           "CLASS' error traceback, make 'debug: True'.",
                           state["params"], self.extra_args)
            raise  # No LoggedError, so that CLASS traceback gets printed
        # Gather products
        for product, collector in self.collectors.items():
            # Special case: sigma8 needs H0, which cannot be known beforehand:
            if "sigma8" in self.collectors:
                self.collectors["sigma8"].args[0] = 8 / self.classy.h()
            method = getattr(self.classy, collector.method)
            arg_array = self.collectors[product].arg_array
            if isinstance(arg_array, int):
                arg_array = np.atleast_1d(arg_array)
            if arg_array is None:
                state[product] = method(
                    *self.collectors[product].args, **self.collectors[product].kwargs)
            elif isinstance(arg_array, Sequence) or isinstance(arg_array, np.ndarray):
                arg_array = np.array(arg_array)
                if len(arg_array.shape) == 1:
                    # if more than one vectorised arg, assume all vectorised in parallel
                    n_values = len(self.collectors[product].args[arg_array[0]])
                    state[product] = np.zeros(n_values)
                    args = deepcopy(list(self.collectors[product].args))
                    for i in range(n_values):
                        for arg_arr_index in arg_array:
                            args[arg_arr_index] = \
                                self.collectors[product].args[arg_arr_index][i]
                        state[product][i] = method(
                            *args, **self.collectors[product].kwargs)
                elif len(arg_array.shape) == 2:
                    if len(arg_array) > 2:
                        raise NotImplementedError("Only 2 array expanded vars so far.")
                    # Create outer combinations
                    x_and_y = np.array(np.meshgrid(
                        self.collectors[product].args[arg_array[0, 0]],
                        self.collectors[product].args[arg_array[1, 0]])).T
                    args = deepcopy(list(self.collectors[product].args))
                    result = np.empty(shape=x_and_y.shape[:2])
                    for i, row in enumerate(x_and_y):
                        for j, column_element in enumerate(x_and_y[i]):
                            args[arg_array[0, 0]] = column_element[0]
                            args[arg_array[1, 0]] = column_element[1]
                            result[i, j] = method(
                                *args, **self.collectors[product].kwargs)
                    state[product] = (
                        self.collectors[product].args[arg_array[0, 0]],
                        self.collectors[product].args[arg_array[1, 0]], result)
                else:
                    raise ValueError("arg_array not correctly formatted.")
            elif arg_array in self.collectors[product].kwargs:
                value = np.atleast_1d(self.collectors[product].kwargs[arg_array])
                state[product] = np.zeros(value.shape)
                for i, v in enumerate(value):
                    kwargs = deepcopy(self.collectors[product].kwargs)
                    kwargs[arg_array] = v
                    state[product][i] = method(
                        *self.collectors[product].args, **kwargs)
            else:
                raise LoggedError(self.log, "Variable over which to do an array call "
                                            f"not known: arg_array={arg_array}")
            if collector.post:
                state[product] = collector.post(*state[product])
        # Prepare derived parameters
        d, d_extra = self._get_derived_all(derived_requested=want_derived)
        if want_derived:
            state["derived"] = {p: d.get(p) for p in self.output_params}
            # Prepare necessary extra derived parameters
        state["derived_extra"] = deepcopy(d_extra)

    def _get_derived_all(self, derived_requested=True):
        """
        Returns a dictionary of derived parameters with their values,
        using the *current* state (i.e. it should only be called from
        the ``compute`` method).

        Parameter names are returned in CLASS nomenclature.

        To get a parameter *from a likelihood* use `get_param` instead.
        """
        # TODO: fails with derived_requested=False
        # Put all parameters in CLASS nomenclature (self.derived_extra already is)
        requested = [self.translate_param(p) for p in (
            self.output_params if derived_requested else [])]
        requested_and_extra = dict.fromkeys(set(requested).union(self.derived_extra))
        # Parameters with their own getters
        if "rs_drag" in requested_and_extra:
            requested_and_extra["rs_drag"] = self.classy.rs_drag()
        if "Omega_nu" in requested_and_extra:
            requested_and_extra["Omega_nu"] = self.classy.Omega_nu
        if "T_cmb" in requested_and_extra:
            requested_and_extra["T_cmb"] = self.classy.T_cmb()
        # Get the rest using the general derived param getter
        # No need for error control: classy.get_current_derived_parameters is passed
        # every derived parameter not excluded before, and cause an error, indicating
        # which parameters are not recognized
        requested_and_extra.update(
            self.classy.get_current_derived_parameters(
                [p for p, v in requested_and_extra.items() if v is None]))
        # Separate the parameters before returning
        # Remember: self.output_params is in sampler nomenclature,
        # but self.derived_extra is in CLASS
        derived = {
            p: requested_and_extra[self.translate_param(p)] for p in self.output_params}
        derived_extra = {p: requested_and_extra[p] for p in self.derived_extra}
        return derived, derived_extra

    def _get_Cl(self, ell_factor=False, units="FIRASmuK2", lensed=True):
        which_key = "Cl" if lensed else "unlensed_Cl"
        which_error = "lensed" if lensed else "unlensed"
        try:
            cls = deepcopy(self.current_state[which_key])
        except:
            raise LoggedError(self.log, "No %s Cl's were computed. Are you sure that you "
                                        "have requested them?", which_error)
        # unit conversion and ell_factor
        ells_factor = \
            ((cls["ell"] + 1) * cls["ell"] / (2 * np.pi))[2:] if ell_factor else 1
        units_factor = self._cmb_unit_factor(
            units, self.current_state['derived_extra']['T_cmb'])
        for cl in cls:
            if cl not in ['pp', 'ell']:
                cls[cl][2:] *= units_factor ** 2 * ells_factor
        if lensed and "pp" in cls and ell_factor:
            cls['pp'][2:] *= ells_factor ** 2 * (2 * np.pi)
        return cls

    def get_Cl(self, ell_factor=False, units="FIRASmuK2"):
        return self._get_Cl(ell_factor=ell_factor, units=units, lensed=True)

    def get_unlensed_Cl(self, ell_factor=False, units="FIRASmuK2"):
        return self._get_Cl(ell_factor=ell_factor, units=units, lensed=False)

    def get_CLASS_background(self):
        """Direct access to ``get_background`` from the CLASS python interface."""
        return self.current_state["CLASS_background"]

    def get_CLASS_thermodynamics(self):
        """Direct access to ``get_thermodynamics`` from the CLASS python interface."""
        return self.current_state["CLASS_thermodynamics"]

    def get_CLASS_primordial(self):
        """Direct access to ``get_primordial`` from the CLASS python interface."""
        return self.current_state["CLASS_primordial"]

    def get_CLASS_perturbations(self):
        """Direct access to ``get_perturbations`` from the CLASS python interface."""
        return self.current_state["CLASS_perturbations"]

    def get_CLASS_sources(self):
        """Direct access to ``get_sources`` from the CLASS python interface."""
        return self.current_state["CLASS_sources"]

    def close(self):
        self.classy.empty()

    def get_can_provide_params(self):
        names = ['Omega_Lambda', 'Omega_cdm', 'Omega_b', 'Omega_m', 'rs_drag', 'z_reio',
                 'YHe', 'Omega_k', 'age', 'sigma8']
        for name, mapped in self.renames.items():
            if mapped in names:
                names.append(name)
        return names

    def get_can_support_params(self):
        # non-exhaustive list of supported input parameters that will be assigned to
        # classy if they are varied
        return ['H0']

    def get_version(self):
        return getattr(self.classy_module, '__version__', None)

    # Installation routines

    @classmethod
    def get_path(cls, path):
        return os.path.realpath(os.path.join(path, "code", cls.__name__))

    @staticmethod
    def get_import_path(path):
        return get_compiled_import_path(os.path.join(path, "python"))

    @classmethod
    def is_compatible(cls):
        import platform
        if platform.system() == "Windows":
            return False
        return True

    @classmethod
    def is_installed(cls, reload=False, **kwargs):
        if not kwargs.get("code", True):
            return True
        try:
            return bool(load_external_module(
                "classy", path=kwargs["path"], get_import_path=cls.get_import_path,
                min_version=cls._classy_repo_version, reload=reload,
                logger=get_logger(cls.__name__), not_installed_level="debug"))
        except ComponentNotInstalledError:
            return False

    @classmethod
    def install(cls, path=None, code=True, no_progress_bars=False, **_kwargs):
        log = get_logger(cls.__name__)
        if not code:
            log.info("Code not requested. Nothing to do.")
            return True
        log.info("Installing pre-requisites...")
        exit_status = pip_install("cython")
        if exit_status:
            log.error("Could not install pre-requisite: cython")
            return False
        log.info("Downloading classy...")
        success = download_github_release(
            os.path.join(path, "code"), cls._classy_repo_name, cls._classy_repo_version,
            directory=cls.__name__, no_progress_bars=no_progress_bars, logger=log)
        if not success:
            log.error("Could not download classy.")
            return False
        # Compilation
        # gcc check after downloading, in case the user wants to change the compiler by
        # hand in the Makefile
        classy_path = cls.get_path(path)
        if not check_gcc_version(cls._classy_min_gcc_version, error_returns=False):
            log.error("Your gcc version is too low! CLASS would probably compile, "
                      "but it would leak memory when running a chain. Please use a "
                      "gcc version newer than %s. You can still compile CLASS by hand, "
                      "maybe changing the compiler in the Makefile. CLASS has been "
                      "downloaded into %r",
                      cls._classy_min_gcc_version, classy_path)
            return False
        log.info("Compiling classy...")
        from subprocess import Popen, PIPE
        env = deepcopy(os.environ)
        env.update({"PYTHON": sys.executable})
        process_make = Popen(["make"], cwd=classy_path, stdout=PIPE, stderr=PIPE, env=env)
        out, err = process_make.communicate()
        if process_make.returncode:
            log.info(out)
            log.info(err)
            log.error("Compilation failed!")
            return False
        return True



"""
.. module:: DataSetLikelihood

:Synopsis: Base class for .dataset based likelihoods.
:Author: Antony Lewis

"""
# Global
import os
from getdist import IniFile

# Local
from cobaya.log import LoggedError
from cobaya.component import ComponentNotInstalledError
from cobaya.typing import InfoDict
from cobaya.conventions import packages_path_input
from .InstallableLikelihood import InstallableLikelihood


class _fast_chi_square:
    def __get__(self, instance, owner):
        # delay testing active camb until run time
        try:
            from camb.mathutils import chi_squared as fast_chi_squared
        except ImportError:
            def fast_chi_squared(covinv, x):
                return covinv.dot(x).dot(x)

        instance.fast_chi_squared = fast_chi_squared
        return fast_chi_squared


class DataSetLikelihood(InstallableLikelihood):
    """A likelihood reading parameters and file names from a .dataset plain text
    .ini file (as CosmoMC)"""

    _default_dataset_params: InfoDict = {}

    _fast_chi_squared = _fast_chi_square()

    # variables defined in yaml or input dictionary
    dataset_file: str

    def initialize(self):

        if os.path.isabs(self.dataset_file):
            data_file = self.dataset_file
            self.path = os.path.dirname(data_file)
        else:
            # If no path specified and has install options (so it installed its data as an
            # external package), use the external packages path
            if not self.path and self.get_install_options() and self.packages_path:
                self.path = self.get_path(self.packages_path)
            self.path = self.path or self.get_class_path()
            if not self.path:
                raise LoggedError(self.log,
                                  "No path given for %s. Set the likelihood "
                                  "property 'path' or the common property '%s'.",
                                  self.dataset_file, packages_path_input)

            data_file = os.path.normpath(os.path.join(self.path, self.dataset_file))
        if not os.path.exists(data_file):
            raise ComponentNotInstalledError(
                self.log, "The data file '%s' could not be found at '%s'. "
                          "Either you have not installed this likelihood, "
                          "or have given the wrong packages installation path.",
                self.dataset_file, self.path)
        self.load_dataset_file(data_file, getattr(self, 'dataset_params', {}))

    def load_dataset_file(self, filename, dataset_params=None):
        if '.dataset' not in filename:
            filename += '.dataset'
        ini = IniFile(filename)
        self.dataset_filename = filename
        ini.params.update(self._default_dataset_params)
        ini.params.update(dataset_params or {})
        self.init_params(ini)

    def init_params(self, ini):
        assert False, "init_params should be inherited"



from . import batchjob_args, jobqueue


# TODO: currently only tested for use with runMPI.py for single jobs

def running_jobs():
    opts = batchjob_args.BatchArgs(
        'List details of running or queued jobs; gives job stats, '
        'then current R-1 and job/chain names', importance=True, batchPathOptional=True)

    group = opts.parser.add_mutually_exclusive_group()
    group.add_argument('--queued', action='store_true')
    group.add_argument('--running', action='store_true')

    (batch, args) = opts.parseForBatch()

    if batch:
        items = [jobItem for jobItem in opts.filteredBatchItems()]
        batch_names = set(
            [jobItem.name for jobItem in items] + [jobItem.name + '_minimize' for jobItem
                                                   in items])
    else:
        items = None
        batch_names = set()

    ids, job_names, nameslist, infos = jobqueue.queue_job_details(args.batchPath,
                                                                  running=not args.queued,
                                                                  queued=not args.running)
    for job_id, job_name, names, info in zip(ids, job_names, nameslist, infos):
        if batch_names.intersection(names) or items is None:
            stats = dict()
            if items:
                for name in names:
                    for jobItem in items:
                        if jobItem.name == name:
                            R = jobItem.convergeStat()[0]
                            if R:
                                stats[name] = "%6.3f" % R
                            break
            R = stats.get(job_name) or ' ' * 6
            print(info + ' |', R, job_name)
            if len(names) > 1:
                for name in names:
                    R = stats.get(name) or ' ' * 6
                    print('    >> ', R, name)


if __name__ == "__main__":
    running_jobs()



import os
from getdist import IniFile, ParamNames
from getdist.parampriors import ParamBounds
from cobaya.typing import InputDict, ParamDict, ParamsDict, LikesDict


def cosmomc_root_to_cobaya_info_dict(root: str, derived_to_input=()) -> InputDict:
    """
    Given the root name of existing cosmomc chain files, tries to construct a Cobaya
    input parameter dictionary with roughly equivalent settings. The output
    dictionary can be used for importance sampling from CosmoMC chains in simple cases
    using Cobaya's 'post'.

    Parameters in the optional derived_to_input list are converted from being derived
    parameters in CosmoMC to non-derived in Cobaya.

    This is by no means guaranteed to produce valid or equivalent results, use at your
    own risk with careful checking! Note that the parameter dictionary will not have
    settings for CAMB, samplers etc, which is OK for importance sampling but you would
    need to add them as necessary to reproduce results.

    Parameter chains in CosmoMC format are available for Planck
    from https://pla.esac.esa.int/pla/#home

    """
    names = ParamNames(root + '.paramnames')
    if os.path.exists(root + '.ranges'):
        ranges = ParamBounds(root + '.ranges')
    else:
        ranges = None
    d: ParamsDict = {}
    info: InputDict = {'params': d}
    for par, name in zip(names.names, names.list()):
        if name.startswith('chi2_') and not name.startswith('chi2__'):
            if name == 'chi2_prior':
                continue
            name = name.replace('chi2_', 'chi2__')
        if name.startswith('minuslogprior') or name == 'chi2':
            continue
        param_dict: ParamDict = {'latex': par.label}
        d[name] = param_dict
        if par.renames:
            param_dict['renames'] = par.renames
        if par.isDerived:
            if name not in derived_to_input:
                param_dict['derived'] = True
            else:
                par.isDerived = False
        if ranges and name in ranges.names:
            if par.isDerived:
                low_up = ranges.getLower(name), ranges.getUpper(name)
                if any(r is not None for r in low_up):
                    param_dict['min'], param_dict['max'] = low_up
            else:
                param_dict["prior"] = [ranges.getLower(name), ranges.getUpper(name)]
    if ranges:
        d.update(ranges.fixedValueDict())
    if names.numberOfName('As') == -1 and names.numberOfName('logA') != -1:
        d['As'] = {'latex': r'A_\mathrm{s}', 'value': 'lambda logA: 1e-10*np.exp(logA)'}
    if names.numberOfName('cosmomc_theta') == -1 and names.numberOfName('theta') != -1:
        d['cosmomc_theta'] = {'latex': r'\theta_{\rm MC}',
                              'value': 'lambda theta: theta/100'}

    # special case for CosmoMC (e.g. Planck) chains
    if os.path.exists(root + '.inputparams'):
        inputs = IniFile(root + '.inputparams')
        for key, value in inputs.params.items():
            if key.startswith('prior['):
                if 'prior' not in info:
                    info['prior'] = {}
                param = key[6:-1]
                if param in d:
                    mean, std = (float(v.strip()) for v in value.split())
                    if not names.parWithName(param).isDerived:
                        info['prior'][param + '_prior'] = \
                            "lambda %s: stats.norm.logpdf(%s, loc=%g, scale=%g)" % (
                                param, param, mean, std)

            if key.startswith('linear_combination['):
                param = key.replace('linear_combination[', '')[:-1]
                prior = inputs.params.get('prior[%s]' % param, None)
                if prior:
                    weights = inputs.params.get('linear_combination_weights[%s]' % param,
                                                None)
                    if not weights:
                        raise ValueError(
                            'linear_combination[%s] prior found but not weights' % param)
                    weights = [float(w.strip()) for w in weights.split()]
                    inputs = value.split()
                    if 'prior' not in info:
                        info['prior'] = {}
                    mean, std = (float(v.strip()) for v in prior.split())
                    linear = "".join(
                        "%+g*%s" % (_w, _p) for _w, _p in zip(weights, inputs))
                    info['prior']['SZ'] = \
                        "lambda %s: stats.norm.logpdf(%s, loc=%g, scale=%g)" % (
                            ",".join(inputs), linear, mean, std)
    if os.path.exists(root + '.likelihoods'):
        info_like: LikesDict = {}
        info['likelihood'] = info_like
        with open(root + '.likelihoods', 'r') as f:
            for line in f.readlines():
                if line.strip():
                    like = line.split()[2]
                    info_like[like] = None
    else:
        print('You need to mention in the likelihood block with with "name: None"'
              'for each likelihood in the input chain')

    info['output'] = root
    return info



"""
.. module:: samplers.polychord

:Synopsis: Interface for the PolyChord nested sampler
:Author: Will Handley, Mike Hobson and Anthony Lasenby (for PolyChord),
         Jesus Torrado (for the cobaya wrapper only)
"""
# Global
import os
import sys
import logging
import inspect
from itertools import chain
from typing import Any, Callable
from tempfile import gettempdir
import re
import numpy as np

# Local
from cobaya.tools import read_dnumber, get_external_function, find_with_regexp, \
    NumberWithUnits, get_compiled_import_path
from cobaya.sampler import Sampler
from cobaya.mpi import is_main_process, share_mpi, sync_processes
from cobaya.collection import SampleCollection
from cobaya.log import get_logger, NoLogging, LoggedError
from cobaya.install import download_github_release
from cobaya.component import ComponentNotInstalledError, load_external_module
from cobaya.yaml import yaml_dump_file
from cobaya.conventions import derived_par_name_separator, Extension


# Suppresses warnings about first defining attrs outside __init__
# pylint: disable=attribute-defined-outside-init


class polychord(Sampler):
    r"""
    PolyChord sampler \cite{Handley:2015fda,2015MNRAS.453.4384H}, a nested sampler
    tailored for high-dimensional parameter spaces with a speed hierarchy.
    """

    # Name of the PolyChord repo and version to download
    _pc_repo_name = "PolyChord/PolyChordLite"
    _pc_repo_version = "1.20.1"
    _base_dir_suffix = "polychord_raw"
    _clusters_dir = "clusters"
    _at_resume_prefer_old = Sampler._at_resume_prefer_old + ["blocking"]
    _at_resume_prefer_new = Sampler._at_resume_prefer_new + ["callback_function"]
    pypolychord: Any

    # variables from yaml
    do_clustering: bool
    num_repeats: int
    confidence_for_unbounded: float
    callback_function: Callable
    blocking: Any
    measure_speeds: bool
    oversample_power: float
    nlive: NumberWithUnits
    path: str
    logzero: float
    max_ndead: int

    def initialize(self):
        """Imports the PolyChord sampler and prepares its arguments."""
        install_path = self.get_path(self.packages_path) if self.packages_path else None
        try:
            self.pc = load_external_module(
                "pypolychord", path=self.path, install_path=install_path,
                min_version=self._pc_repo_version,
                get_import_path=get_compiled_import_path, logger=self.log,
                not_installed_level="debug")
        except ComponentNotInstalledError as excpt:
            raise ComponentNotInstalledError(
                self.log, (f"Could not find PolyChord: {excpt}. "
                           "To install it, run `cobaya-install polychord`")) from excpt
        with NoLogging(logging.CRITICAL):
            settings = load_external_module(
                "pypolychord.settings", path=self.path, install_path=install_path,
                get_import_path=get_compiled_import_path, logger=self.log)
        # Prepare arguments and settings
        self.n_sampled = len(self.model.parameterization.sampled_params())
        self.n_derived = len(self.model.parameterization.derived_params())
        self.n_priors = len(self.model.prior)
        self.n_likes = len(self.model.likelihood)
        self.nDims = self.model.prior.d()
        self.nDerived = self.n_derived + self.n_priors + self.n_likes
        if self.logzero is None:
            self.logzero = np.nan_to_num(-np.inf)
        if self.max_ndead == np.inf:
            self.max_ndead = -1
        self._quants_d_units = ["nlive", "max_ndead"]
        for p in self._quants_d_units:
            if getattr(self, p) is not None:
                setattr(self, p, NumberWithUnits(
                    getattr(self, p), "d", scale=self.nDims, dtype=int).value)
        self._quants_nlive_units = ["nprior", "nfail"]
        for p in self._quants_nlive_units:
            if getattr(self, p) is not None:
                setattr(self, p, NumberWithUnits(
                    getattr(self, p), "nlive", scale=self.nlive, dtype=int).value)
        # Fill the automatic ones
        if getattr(self, "feedback", None) is None:
            values = {logging.CRITICAL: 0, logging.ERROR: 0, logging.WARNING: 0,
                      logging.INFO: 1, logging.DEBUG: 2}
            self.feedback = values[self.log.getEffectiveLevel()]
        # Prepare output folders and prefixes
        if self.output:
            self.file_root = self.output.prefix
            self.read_resume = self.output.is_resuming()
        else:
            output_prefix = share_mpi(hex(int(self._rng.random() * 16 ** 6))[2:]
                                      if is_main_process() else None)
            self.file_root = output_prefix
            # dummy output -- no resume!
            self.read_resume = False
        self.base_dir = self.get_base_dir(self.output)
        self.raw_clusters_dir = os.path.join(self.base_dir, self._clusters_dir)
        self.output.create_folder(self.base_dir)
        if self.do_clustering:
            self.clusters_folder = self.get_clusters_dir(self.output)
            self.output.create_folder(self.clusters_folder)
        self.mpi_info("Storing raw PolyChord output in '%s'.", self.base_dir)
        # Exploiting the speed hierarchy
        if self.blocking:
            blocks, oversampling_factors = self.model.check_blocking(self.blocking)
        else:
            if self.measure_speeds:
                self.model.measure_and_set_speeds(n=self.measure_speeds,
                                                  random_state=self._rng)
            blocks, oversampling_factors = self.model.get_param_blocking_for_sampler(
                oversample_power=self.oversample_power)
        self.mpi_info("Parameter blocks and their oversampling factors:")
        max_width = len(str(max(oversampling_factors)))
        for f, b in zip(oversampling_factors, blocks):
            self.mpi_info("* %" + "%d" % max_width + "d : %r", f, b)
        # Save blocking in updated info, in case we want to resume
        self._updated_info["blocking"] = list(zip(oversampling_factors, blocks))
        blocks_flat = list(chain(*blocks))
        self.ordering = [
            blocks_flat.index(p) for p in self.model.parameterization.sampled_params()]
        self.grade_dims = [len(block) for block in blocks]
        # Steps per block
        # NB: num_repeats is ignored by PolyChord when int "grade_frac" given,
        # so needs to be applied by hand.
        # In num_repeats, `d` is interpreted as dimension of each block
        self.grade_frac = [
            int(o * read_dnumber(self.num_repeats, dim_block))
            for o, dim_block in zip(oversampling_factors, self.grade_dims)]
        # Assign settings
        pc_args = ["nlive", "num_repeats", "nprior", "nfail", "do_clustering",
                   "feedback", "precision_criterion", "logzero",
                   "max_ndead", "boost_posterior", "posteriors", "equals",
                   "cluster_posteriors", "write_resume", "read_resume",
                   "write_stats", "write_live", "write_dead", "write_prior",
                   "maximise", "compression_factor", "synchronous", "base_dir",
                   "file_root", "grade_dims", "grade_frac", "nlives"]
        # As stated above, num_repeats is ignored, so let's not pass it
        pc_args.pop(pc_args.index("num_repeats"))
        self.pc_settings = settings.PolyChordSettings(
            self.nDims, self.nDerived, seed=(self.seed if self.seed is not None else -1),
            **{p: getattr(self, p) for p in pc_args if getattr(self, p) is not None})
        # prior conversion from the hypercube
        bounds = self.model.prior.bounds(
            confidence_for_unbounded=self.confidence_for_unbounded)
        # Check if priors are bounded (nan's to inf)
        inf = np.where(np.isinf(bounds))
        if len(inf[0]):
            params_names = list(self.model.parameterization.sampled_params())
            params = [params_names[i] for i in sorted(set(inf[0]))]
            raise LoggedError(
                self.log, "PolyChord needs bounded priors, but the parameter(s) '"
                          "', '".join(params) + "' is(are) unbounded.")
        locs = bounds[:, 0]
        scales = bounds[:, 1] - bounds[:, 0]
        # This function re-scales the parameters AND puts them in the right order
        self.pc_prior = lambda x: (locs + np.array(x)[self.ordering] * scales).tolist()
        # We will need the volume of the prior domain, since PolyChord divides by it
        self.logvolume = np.log(np.prod(scales))
        # Prepare callback function
        if self.callback_function is not None:
            self.callback_function_callable = (
                get_external_function(self.callback_function))
        self.last_point_callback = 0
        # Prepare runtime live and dead points collections
        self.live = SampleCollection(self.model, None, name="live")
        self.dead = SampleCollection(self.model, self.output, name="dead")
        # Done!
        if is_main_process():
            self.log.debug("Calling PolyChord with arguments:")
            for p, v in inspect.getmembers(self.pc_settings, lambda a: not callable(a)):
                if not p.startswith("_"):
                    self.log.debug("  %s: %s", p, v)
        self.logZ, self.logZstd = np.nan, np.nan
        self._frac_unphysical = np.nan
        self.collection = None
        self.clusters = None

    def dumper(self, live_points, dead_points, logweights, logZ, logZstd):
        """
        Preprocess output for the callback function and calls it, if present.
        """
        if self.callback_function is None:
            return
        # Store live and dead points and evidence computed so far
        self.live.reset()
        for point in live_points:
            self.live.add(
                point[:self.n_sampled],
                derived=point[self.n_sampled:self.n_sampled + self.n_derived],
                weight=np.nan,
                logpriors=point[self.n_sampled + self.n_derived:
                                self.n_sampled + self.n_derived + self.n_priors],
                loglikes=point[self.n_sampled + self.n_derived + self.n_priors:
                               self.n_sampled + self.n_derived + self.n_priors +
                               self.n_likes])
        for logweight, point in zip(logweights[self.last_point_callback:],
                                    dead_points[self.last_point_callback:]):
            self.dead.add(
                point[:self.n_sampled],
                derived=point[self.n_sampled:self.n_sampled + self.n_derived],
                weight=np.exp(logweight),
                logpriors=point[self.n_sampled + self.n_derived:
                                self.n_sampled + self.n_derived + self.n_priors],
                loglikes=point[self.n_sampled + self.n_derived + self.n_priors:
                               self.n_sampled + self.n_derived + self.n_priors +
                               self.n_likes])
        self.logZ, self.logZstd = logZ, logZstd
        self._correct_unphysical_fraction()
        # Callback function
        if self.callback_function is not None:
            try:
                self.callback_function_callable(self)
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.log.error("The callback function produced an error: %r", str(e))
            self.last_point_callback = len(self.dead)

    def run(self):
        """
        Prepares the posterior function and calls ``PolyChord``'s ``run`` function.
        """

        # Prepare the posterior
        # Don't forget to multiply by the volume of the physical hypercube,
        # since PolyChord divides by it
        def logpost(params_values):
            result = self.model.logposterior(params_values)
            loglikes = result.loglikes
            if len(loglikes) != self.n_likes:
                loglikes = np.full(self.n_likes, np.nan)
            derived = result.derived
            if len(derived) != self.n_derived:
                derived = np.full(self.n_derived, np.nan)
            derived = list(derived) + list(result.logpriors) + list(loglikes)
            return (max(result.logpost + self.logvolume, self.pc_settings.logzero),
                    derived)

        sync_processes()
        self.mpi_info("Calling PolyChord...")
        self.pc.run_polychord(logpost, self.nDims, self.nDerived, self.pc_settings,
                              self.pc_prior, self.dumper)
        self.process_raw_output()

    @property
    def raw_prefix(self):
        return os.path.join(
            self.pc_settings.base_dir, self.pc_settings.file_root)

    def dump_paramnames(self, prefix):
        labels = self.model.parameterization.labels()
        with open(prefix + ".paramnames", "w", encoding="utf-8-sig") as f_paramnames:
            for p in self.model.parameterization.sampled_params():
                f_paramnames.write("%s\t%s\n" % (p, labels.get(p, "")))
            for p in self.model.parameterization.derived_params():
                f_paramnames.write("%s*\t%s\n" % (p, labels.get(p, "")))
            for p in self.model.prior:
                f_paramnames.write("%s*\t%s\n" % (
                    "logprior" + derived_par_name_separator + p,
                    r"\log\pi_\mathrm{" + p.replace("_", r"\ ") + r"}"))
            for p in self.model.likelihood:
                f_paramnames.write("%s*\t%s\n" % (
                    "loglike" + derived_par_name_separator + p,
                    r"\log\mathcal{L}_\mathrm{" + p.replace("_", r"\ ") + r"}"))

    def save_sample(self, fname, name):
        sample = np.atleast_2d(np.loadtxt(fname))
        if not sample.size:
            return None
        collection = SampleCollection(
            self.model, self.output, name=str(name), sample_type="nested")
        for row in sample:
            collection.add(
                row[2:2 + self.n_sampled],
                derived=row[2 + self.n_sampled:2 + self.n_sampled + self.n_derived],
                weight=row[0],
                logpriors=row[-(self.n_priors + self.n_likes):-self.n_likes],
                loglikes=row[-self.n_likes:])
        # make sure that the points are written
        collection.out_update()
        return collection

    def _correct_unphysical_fraction(self):
        """
        Correction for the fraction of the prior that is unphysical -- see issue #77
        """
        if np.isnan(self._frac_unphysical):
            with open(self.raw_prefix + ".prior_info", "r", encoding="utf-8-sig") as pf:
                lines = list(pf.readlines())
            get_value_str = lambda line: line[line.find("=") + 1:]
            get_value_str_var = lambda var: get_value_str(
                next(line for line in lines if line.lstrip().startswith(var)))
            nprior = int(get_value_str_var("nprior"))
            ndiscarded = int(get_value_str_var("ndiscarded"))
            self._frac_unphysical = nprior / ndiscarded
        if self._frac_unphysical != 1:
            self.log.debug(
                "Correcting for unphysical region fraction: %g", self._frac_unphysical)
            self.logZ += np.log(self._frac_unphysical)
            if self.clusters is not None:
                for cluster in self.clusters.values():
                    cluster["logZ"] += np.log(self._frac_unphysical)

    def process_raw_output(self):
        """
        Loads the sample of live points from ``PolyChord``'s raw output and writes it
        (if ``txt`` output requested).
        """
        if is_main_process():
            self.log.info("Loading PolyChord's results: samples and evidences.")
            self.dump_paramnames(self.raw_prefix)
            self.collection = self.save_sample(self.raw_prefix + ".txt", "1")
            # Load clusters, and save if output
            if self.pc_settings.do_clustering:
                self.clusters = {}
                clusters_raw_regexp = re.compile(
                    re.escape(self.pc_settings.file_root + "_") + r"\d+\.txt")
                cluster_raw_files = sorted(find_with_regexp(
                    clusters_raw_regexp, os.path.join(
                        self.pc_settings.base_dir, self._clusters_dir), walk_tree=True))
                for f in cluster_raw_files:
                    i = int(f[f.rfind("_") + 1:-len(".txt")])
                    if self.output:
                        old_folder = self.output.folder
                        self.output.folder = self.clusters_folder
                    sample = self.save_sample(f, str(i))
                    if self.output:
                        # noinspection PyUnboundLocalVariable
                        self.output.folder = old_folder
                    self.clusters[i] = {"sample": sample}
            # Prepare the evidence(s) and write to file
            pre = "log(Z"
            active = "(Still active)"
            with open(self.raw_prefix + ".stats", "r", encoding="utf-8-sig") as statsfile:
                lines = [line for line in statsfile.readlines() if line.startswith(pre)]
            for line in lines:
                logZ, logZstd = [float(n.replace(active, "")) for n in
                                 line.split("=")[-1].split("+/-")]
                component = line.split("=")[0].lstrip(pre + "_").rstrip(") ")
                if not component:
                    self.logZ, self.logZstd = logZ, logZstd
                elif self.pc_settings.do_clustering:
                    i = int(component)
                    self.clusters[i]["logZ"], self.clusters[i]["logZstd"] = logZ, logZstd
            self.log.debug(
                "RAW log(Z) = %g +/- %g ; RAW Z in [%.8g, %.8g] (68%% C.L. log-gaussian)",
                self.logZ, self.logZstd,
                *[np.exp(self.logZ + n * self.logZstd) for n in [-1, 1]])
            self._correct_unphysical_fraction()
            if self.output:
                out_evidences = {"logZ": self.logZ, "logZstd": self.logZstd}
                if self.clusters is not None:
                    out_evidences["clusters"] = {}
                    for i in sorted(list(self.clusters)):
                        out_evidences["clusters"][i] = {
                            "logZ": self.clusters[i]["logZ"],
                            "logZstd": self.clusters[i]["logZstd"],
                        }
                fname = os.path.join(self.output.folder,
                                     self.output.prefix + Extension.evidence)
                yaml_dump_file(fname, out_evidences, comment="log-evidence",
                               error_if_exists=False)
        # TODO: try to broadcast the collections
        # if get_mpi():
        #     bcast_from_0 = lambda attrname: setattr(self,
        #         attrname, get_mpi_comm().bcast(getattr(self, attrname, None), root=0))
        #     map(bcast_from_0, ["collection", "logZ", "logZstd", "clusters"])
        if is_main_process():
            self.log.info("Finished! Raw PolyChord output stored in '%s', "
                          "with prefix '%s'",
                          self.pc_settings.base_dir, self.pc_settings.file_root)
            self.log.info(
                "log(Z) = %g +/- %g ; Z in [%.8g, %.8g] (68%% C.L. log-gaussian)",
                self.logZ, self.logZstd,
                *[np.exp(self.logZ + n * self.logZstd) for n in [-1, 1]])

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``SampleCollection`` containing the sequentially
           discarded live points.
        """
        if is_main_process():
            products = {
                "sample": self.collection, "logZ": self.logZ, "logZstd": self.logZstd}
            if self.pc_settings.do_clustering:
                products.update({"clusters": self.clusters})
            return products
        else:
            return {}

    @classmethod
    def get_base_dir(cls, output):
        if output:
            return output.add_suffix(cls._base_dir_suffix, separator="_")
        return os.path.join(gettempdir(), cls._base_dir_suffix)

    @classmethod
    def get_clusters_dir(cls, output):
        if output:
            return output.add_suffix(cls._clusters_dir, separator="_")

    @classmethod
    def output_files_regexps(cls, output, info=None, minimal=False):
        # Resume file
        regexps_tuples = [
            (re.compile(re.escape(output.prefix + ".resume")), cls.get_base_dir(output))]
        if minimal:
            return regexps_tuples
        return regexps_tuples + [
            # Raw products base dir
            (None, cls.get_base_dir(output)),
            # Main sample
            (output.collection_regexp(name=None), None),
            # Evidence
            (re.compile(re.escape(output.prefix + Extension.evidence)), None),
            # Clusters
            (None, cls.get_clusters_dir(output))
        ]

    @classmethod
    def get_version(cls):
        return None

    @classmethod
    def get_path(cls, path):
        return os.path.realpath(
            os.path.join(path, "code",
                         cls._pc_repo_name[cls._pc_repo_name.find("/") + 1:]))

    @classmethod
    def is_compatible(cls):
        import platform  # pylint: disable=import-outside-toplevel
        if platform.system() == "Windows":
            return False
        return True

    @classmethod
    def is_installed(cls, reload=False, **kwargs):
        if not kwargs.get("code", True):
            return True
        try:
            return bool(load_external_module(
                "pypolychord", path=kwargs["path"],
                get_import_path=get_compiled_import_path,
                min_version=cls._pc_repo_version, reload=reload,
                logger=get_logger(cls.__name__), not_installed_level="debug"))
        except ComponentNotInstalledError:
            return False

    @classmethod
    def install(cls, path=None, code=False, no_progress_bars=False, **_kwargs):
        if not code:
            return True
        log = get_logger(__name__)
        log.info("Downloading PolyChord...")
        success = download_github_release(os.path.join(path, "code"), cls._pc_repo_name,
                                          cls._pc_repo_version,
                                          no_progress_bars=no_progress_bars,
                                          logger=log)
        if not success:
            log.error("Could not download PolyChord.")
            return False
        log.info("Compiling (Py)PolyChord...")
        from subprocess import Popen, PIPE  # pylint: disable=import-outside-toplevel
        # Needs to re-define os' PWD,
        # because MakeFile calls it and is not affected by the cwd of Popen
        cwd = os.path.join(path, "code",
                           cls._pc_repo_name[cls._pc_repo_name.find("/") + 1:])
        my_env = os.environ.copy()
        my_env.update({"PWD": cwd})
        if "CC" not in my_env:
            my_env["CC"] = "mpicc"
        if "CXX" not in my_env:
            my_env["CXX"] = "mpicxx"
        process_make = Popen([sys.executable, "setup.py", "build"],
                             cwd=cwd, env=my_env, stdout=PIPE, stderr=PIPE)
        out, err = process_make.communicate()
        if process_make.returncode:
            log.info(out.decode("utf-8"))
            log.info(err.decode("utf-8"))
            log.error("Python build failed!")
            return False
        return True



#!/usr/bin/env python

import hashlib
import os

from cobaya.grid_tools import batchjob_args, jobqueue
from cobaya.tools import warn_deprecation


def run():
    warn_deprecation()
    Opts = batchjob_args.BatchArgs(
        prog='cobaya grid-run',
        desc='Submit jobs to run chains or importance sample',
        notExist=True, notall=True, converge=True)
    jobqueue.addArguments(Opts.parser, combinedJobs=True)
    Opts.parser.add_argument('--subitems', action='store_true',
                             help='include sub-grid items')
    Opts.parser.add_argument('--not_queued', action='store_true')
    Opts.parser.add_argument('--Minimize', action='store_true',
                             help='Run minimization jobs')
    Opts.parser.add_argument('--importance_minimize', action='store_true',
                             help=('Run minimization jobs for chains '
                                   'that are importance sampled'))
    Opts.parser.add_argument('--minimize_failed', action='store_true',
                             help='run where minimization previously failed')
    Opts.parser.add_argument('--checkpoint_run', nargs='?', default=None, const=0,
                             type=float,
                             help=(
                                 'run if stopped and not finished; if optional value '
                                 'given then only run chains with convergence worse than '
                                 'the given value'))
    Opts.parser.add_argument('--importance_ready', action='store_true',
                             help='where parent chain has converged and stopped')
    Opts.parser.add_argument('--importance_changed', action='store_true',
                             help=('run importance jobs where the parent chain has '
                                   'changed since last run'))
    Opts.parser.add_argument('--parent_converge', type=float, default=0,
                             help='minimum R-1 convergence for importance job parent')
    Opts.parser.add_argument('--parent_stopped', action='store_true',
                             help='only run if parent chain is not still running')
    (batch, args) = Opts.parseForBatch()
    if args.not_queued:
        print('Getting queued names...')
        queued = jobqueue.queue_job_names(args.batchPath)

    def notQueued(name):
        for job in queued:
            if name in job:
                # print 'Already running:', name
                return False
        return True

    variant = ''
    if args.importance_minimize:
        variant = '_minimize'
        if args.importance is None:
            args.importance = []
    if args.minimize:
        args.noimportance = True
        variant = '_minimize'
    if args.importance is None:
        if args.importance_changed or args.importance_ready:
            args.importance = []
        else:
            args.noimportance = True
    isMinimize = args.importance_minimize or args.minimize
    if args.combine_one_job_name:
        print('Combining multiple (hopefully fast) into single job script: ' +
              args.combine_one_job_name)
    iniFiles = []
    jobqueue.check_arguments(**args.__dict__)

    def jobName():
        s = "-".join([os.path.splitext(os.path.basename(ini))[0] for ini in iniFiles])
        if len(iniFiles) < 2 or len(s) < 70:
            return s
        base = os.path.basename(iniFiles[0])
        if len(base) > 70:
            base = base[:70]
        return base + '__' + hashlib.md5(s.encode('utf8')).hexdigest()[:16]

    def submitJob(ini):
        if not args.dryrun:
            print('Submitting...' + ini)
        else:
            print('... ' + ini)
        iniFiles.append(ini)
        if args.combine_one_job_name:
            return
        if len(iniFiles) >= args.runs_per_job:
            if args.runs_per_job > 1:
                print('--> jobName: ', jobName())
            jobqueue.submitJob(jobName(), iniFiles, **args.__dict__)
            iniFiles.clear()

    for jobItem in Opts.filteredBatchItems(wantSubItems=args.subitems):
        if ((not args.notexist or isMinimize and not jobItem.chainMinimumExists()
             or not isMinimize and not jobItem.chainExists()) and (
                    not args.minimize_failed or not jobItem.chainMinimumConverged())
            and (isMinimize or args.notall is None or not jobItem.allChainExists(
                    args.notall))) \
                and (not isMinimize or getattr(jobItem, 'want_minimize', True)):
            if not args.parent_converge or not jobItem.isImportanceJob or \
                    jobItem.parent.hasConvergeBetterThan(args.parent_converge):
                if args.converge == 0 or not jobItem.hasConvergeBetterThan(
                        args.converge, returnNotExist=True):
                    if args.checkpoint_run is None or jobItem.wantCheckpointContinue(
                            args.checkpoint_run) and jobItem.notRunning():
                        if (not jobItem.isImportanceJob or isMinimize
                                or (args.importance_ready
                                    and jobItem.parent.chainFinished()
                                    or not args.importance_ready
                                    and jobItem.parent.chainExists())
                                and (not args.importance_changed
                                     or jobItem.parentChanged())
                                and (not args.parent_stopped
                                     or jobItem.parent.notRunning())):
                            if not args.not_queued or notQueued(jobItem.name):
                                submitJob(jobItem.iniFile(variant))

    if len(iniFiles) > 0:
        if args.runs_per_job > 1:
            print('--> jobName: ', jobName())
        jobqueue.submitJob(args.combine_one_job_name or jobName(), iniFiles,
                           sequential=args.combine_one_job_name is not None,
                           **args.__dict__)



import subprocess
import os
import numpy as np
import re
import pickle
import time
import shutil
import multiprocessing
from typing import Any
from distutils import spawn

from cobaya.conventions import Extension
from .conventions import script_folder, script_ext, jobid_ext

code_prefix = 'COBAYA'
default_program = 'cobaya-run -r'


def set_default_program(program, env_prefix):
    global code_prefix, default_program
    code_prefix = env_prefix
    default_program = program


def addArguments(parser, combinedJobs=False):
    parser.add_argument('--nodes', type=int)
    parser.add_argument('--chains-per-node', type=int)
    parser.add_argument('--cores-per-node', type=int)
    parser.add_argument('--mem-per-node', type=int, help="Memory in MB per node")
    parser.add_argument('--walltime')
    if combinedJobs:
        parser.add_argument('--combine-one-job-name',
                            help=('run all one after another, under one job submission '
                                  '(good for many fast operations)'))
        parser.add_argument('--runs-per-job', type=int,
                            default=int(os.environ.get(code_prefix + '_runsPerJob', '1')),
                            help=('submit multiple mpi runs at once from each job script '
                                  '(e.g. to get more than one run per node)'))
    parser.add_argument('--job-template',
                        help="template file for the job submission script")
    parser.add_argument('--program',
                        help='actual program to run (default: %s)' % default_program)
    parser.add_argument('--queue', help='name of queue to submit to')
    parser.add_argument('--jobclass', help='any class name of the job')
    parser.add_argument('--qsub', help='option to change qsub command to something else')
    parser.add_argument('--dryrun', action='store_true',
                        help="just test configuration and give summary for checking, "
                             "don't produce or do anything")
    parser.add_argument('--no_sub', action='store_true',
                        help="produce job script but don't actually submit it")


def check_arguments(**kwargs):
    submitJob(None, None, msg=True, **kwargs)


grid_engine_defaults = {
    'PBS': {},
    'MOAB': {'qsub': 'msub', 'qdel': 'canceljob'},
    'SLURM': {'qsub': 'sbatch', 'qdel': 'scancel', 'qstat': 'squeue'}
}


def engine_default(engine, command):
    options = grid_engine_defaults.get(engine)
    if options:
        return options.get(command, command)
    return command


class JobSettings:
    names: list
    jobId: str
    path: str
    onerun: int
    inputFiles: list
    subTime: float

    def __init__(self, jobName, msg=False, **kwargs):
        self.jobName = jobName
        grid_engine = 'PBS'
        for engine, options in grid_engine_defaults.items():
            if 'qsub' in options:
                if spawn.find_executable(options["qsub"]) is not None:
                    grid_engine = engine
                    break
        else:
            if spawn.find_executable("qstat") is not None:
                try:
                    help_info = subprocess.check_output('qstat -help',
                                                        shell=True).decode().strip()
                    if 'OGS/GE' in help_info:
                        grid_engine = 'OGS'  # Open Grid Scheduler, as on StarCluster
                except:
                    pass
        self.job_template = get_defaulted('job_template', 'job_script', **kwargs)
        if not self.job_template or self.job_template == 'job_script' \
                and not os.path.exists(self.job_template):
            raise ValueError("You must provide a script template with '--job-template', "
                             "or export COBAYA_job_template in your .bashrc")
        try:
            with open(self.job_template, 'r', encoding="utf-8-sig") as f:
                template = f.read()
        except IOError:
            raise ValueError("Job template '%s' not found." % self.job_template)
        try:
            cores = multiprocessing.cpu_count()
            if cores > 64:
                # probably shared memory machine, e.g. Cosmos
                cores = 8
        except:
            cores = 8
        self.cores_per_node = get_defaulted('cores_per_node', cores, tp=int,
                                            template=template, **kwargs)
        if cores == 4:
            perNode = 2
        elif cores % 4 == 0:
            perNode = cores // 4
        elif cores > 3 and cores % 3 == 0:
            perNode = cores // 3
        else:
            perNode = 1
        self.chains_per_node = get_defaulted('chains_per_node', perNode, tp=int,
                                             template=template, **kwargs)
        self.nodes = get_defaulted('nodes', max(1, 4 // perNode), tp=int,
                                   template=template, **kwargs)
        self.nchains = self.nodes * self.chains_per_node
        self.runsPerJob = get_defaulted('runsPerJob', 1, tp=int, template=template,
                                        **kwargs)
        # also defaulted at input so should be set here unless called programmatically
        self.omp = self.cores_per_node / (self.chains_per_node * self.runsPerJob)
        if self.omp != np.floor(self.omp):
            raise Exception('Chains must each have equal number of cores')
        if msg:
            print('Job parameters: %i runs of %i chains on %i nodes, '
                  'each node with %i MPI chains, each chain using %i OpenMP cores '
                  '(%i cores per node)' % (
                      self.runsPerJob, self.nchains, self.nodes, self.chains_per_node,
                      self.omp, self.cores_per_node))
        self.mem_per_node = get_defaulted('mem_per_node', 63900, tp=int,
                                          template=template, **kwargs)
        self.walltime = get_defaulted('walltime', '24:00:00', template=template, **kwargs)
        self.program = get_defaulted('program', default_program, template=template,
                                     **kwargs)
        self.queue = get_defaulted('queue', '', template=template, **kwargs)
        self.jobclass = get_defaulted('jobclass', '', template=template, **kwargs)
        self.gridEngine = get_defaulted('GridEngine', grid_engine, template=template,
                                        **kwargs)
        if grid_engine == 'OGS' and os.getenv('SGE_CLUSTER_NAME', '') == 'starcluster':
            self.qsub = 'qsub -pe orte ##NUMSLOTS##'
        else:
            self.qsub = get_defaulted('qsub', engine_default(grid_engine, 'qsub'),
                                      template=template, **kwargs)
        self.qdel = get_defaulted('qdel', engine_default(grid_engine, 'qdel'),
                                  template=template, **kwargs)
        self.runCommand = extract_value(template, 'RUN')


class JobIndex:
    """
    Stores the mappings between job Ids, jobNames
    """

    def __init__(self):
        self.jobSettings = dict()
        self.jobNames = dict()
        self.rootNames = dict()
        self.jobSequence = []

    def addJob(self, j):
        self.jobSettings[j.jobId] = j
        self.jobNames[j.jobName] = j.jobId
        for name in j.names:
            self.rootNames[name] = j.jobId
        self.jobSequence.append(j.jobId)

    def delId(self, jobId):
        if jobId is not None:
            j = self.jobSettings.get(jobId)
            if j is not None:
                for rootname in j.names:
                    del self.rootNames[rootname]
                del self.jobSettings[jobId]
                del self.jobNames[j.jobName]
                self.jobSequence = [s for s in self.jobSequence if s != jobId]


def loadJobIndex(batchPath, must_exist=False):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    fileName = os.path.join(batchPath, 'job_index.pyobj')
    if os.path.exists(fileName):
        with open(fileName, 'rb') as inp:
            return pickle.load(inp)
    else:
        if not must_exist:
            return JobIndex()
        return None


def saveJobIndex(obj, batchPath=None):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    fname = os.path.join(batchPath, 'job_index.pyobj')
    with open(fname + '_tmp', 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)
    # try to prevent corruption from error mid-write
    if os.path.exists(fname):
        os.remove(fname)
    shutil.move(fname + '_tmp', fname)


def addJobIndex(batchPath, j):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    index = loadJobIndex(batchPath)
    index.addJob(j)
    saveJobIndex(index, batchPath)


def deleteJobNames(batchPath, jobNames):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    index = loadJobIndex(batchPath)
    if not index:
        raise Exception('No existing job index found')
    if isinstance(jobNames, str):
        jobNames = [jobNames]
    for name in jobNames:
        jobId = index.jobNames.get(name)
        index.delId(jobId)
    saveJobIndex(index, batchPath)


def deleteRootNames(batchPath, rootNames):
    deleteJobs(batchPath, rootNames=rootNames)


def deleteJobs(batchPath, jobIds=None, rootNames=None, jobNames=None, jobId_minmax=None,
               jobId_min=None, confirm=True, running=False, queued=False):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    index = loadJobIndex(batchPath)
    if not index:
        raise Exception('No existing job index found')
    if jobIds is None:
        jobIds = []
    if isinstance(jobIds, str):
        jobIds = [jobIds]
    jobIds = set(jobIds)
    if rootNames is not None:
        if isinstance(rootNames, str):
            rootNames = [rootNames]
        jobIds.update(index.rootNames.get(name) for name in rootNames)
    if jobNames is not None:
        if isinstance(jobNames, str):
            jobNames = [jobNames]
        jobIds.update(index.jobNames.get(name) for name in jobNames)
    if jobId_minmax is not None or jobId_min is not None:
        for jobIdStr, j in list(index.jobSettings.items()):
            parts = jobIdStr.split('.')
            if len(parts) == 1 or parts[0].isdigit():
                jobId = int(parts[0])
            else:
                jobId = int(parts[1])
            if (jobId_minmax is not None and
                    (jobId_minmax[0] <= jobId <= jobId_minmax[1]) or
                    jobId_min is not None and jobId >= jobId_min):
                jobIds.add(jobIdStr)
    validIds = queue_job_details(batchPath, running=not queued, queued=not running)[0]
    for jobId in sorted(jobIds):
        j = index.jobSettings.get(jobId)
        if j is not None:
            if confirm:
                if jobId in validIds:
                    print('Cancelling: ', j.jobName, jobId)
                    if hasattr(j, 'qdel'):
                        qdel = j.qdel
                    else:
                        qdel = 'qdel'
                    subprocess.check_output(qdel + ' ' + str(jobId), shell=True)
                index.delId(jobId)
            elif jobId in validIds:
                print('...', j.jobName, jobId)
    if confirm:
        saveJobIndex(index, batchPath)
    return jobIds


def submitJob(job_name, input_files, sequential=False, msg=False, **kwargs):
    """
    Submits a job with name ``jobName`` and input file(s) ``input_files``.

    If ``sequential=True`` (default: False), and multiple input files given,
    gathers the corresponding runs in a single job script, so that they are run
    sequentially.

    The ``JobSettings`` are created using the non-explicit ``kwargs``.
    """
    j = JobSettings(job_name, msg, **kwargs)
    if kwargs.get('dryrun', False) or input_files is None:
        return
    if isinstance(input_files, str):
        input_files = [input_files]
    input_files = [
        (os.path.splitext(p)[0] if os.path.splitext(p)[1] in Extension.yamls else p)
        for p in input_files]
    j.runsPerJob = (len(input_files), 1)[sequential]
    # adjust omp for the actual number
    # (may not be equal to input runsPerJob because of non-integer multiple)
    j.omp = j.cores_per_node // (j.chains_per_node * j.runsPerJob)
    j.path = os.getcwd()
    j.onerun = (0, 1)[len(input_files) == 1 or sequential]
    base_path = os.path.abspath(kwargs.get('batchPath', './'))
    script_dir = os.path.join(base_path, script_folder)
    if not os.path.exists(script_dir):
        os.makedirs(script_dir)
    vals = dict()
    vals['JOBNAME'] = job_name
    vals['OMP'] = j.omp
    vals['MEM_MB'] = j.mem_per_node
    vals['WALLTIME'] = j.walltime
    vals['NUMNODES'] = j.nodes
    vals['NUMRUNS'] = j.runsPerJob
    vals['NUMMPI'] = j.nchains
    vals['CHAINSPERNODE'] = j.chains_per_node
    vals['PPN'] = j.chains_per_node * j.runsPerJob * j.omp
    vals['MPIPERNODE'] = j.chains_per_node * j.runsPerJob
    vals['NUMTASKS'] = j.nchains * j.runsPerJob
    vals['NUMSLOTS'] = vals['PPN'] * j.nodes
    vals['ROOTDIR'] = j.path
    vals['ONERUN'] = j.onerun
    vals['PROGRAM'] = j.program
    vals['QUEUE'] = j.queue
    vals['JOBSCRIPTDIR'] = script_dir
    if hasattr(j, 'jobclass'):
        vals['JOBCLASS'] = j.jobclass
    j.names = [os.path.basename(param) for param in input_files]
    commands = []
    for param, name in zip(input_files, j.names):
        ini = param + '.yaml'
        if j.runCommand is not None:
            vals['INI'] = ini
            vals['INIBASE'] = name
            command = replace_placeholders(j.runCommand, vals)
        else:
            command = ('mpirun -np %i %s %s' % (j.nchains, j.program, ini))
        commands.append(command)
    vals['COMMAND'] = "\n".join(commands)
    with open(j.job_template, 'r', encoding="utf-8-sig") as f:
        template = f.read()
        # Remove definition lines
        template = "\n".join(
            [line for line in template.split("\n") if not line.startswith("##")])
        script = replace_placeholders(template, vals)
        scriptRoot = os.path.join(script_dir, job_name)
        scriptName = scriptRoot + script_ext
        open(scriptName, 'w', encoding="utf-8").write(script)
        if len(input_files) > 1:
            open(scriptRoot + '.batch', 'w', encoding="utf-8").write(
                "\n".join(input_files))
        if not kwargs.get('no_sub', False):
            try:
                res = subprocess.check_output(
                    replace_placeholders(j.qsub, vals) + ' ' + scriptName, shell=True,
                    stderr=subprocess.STDOUT).decode().strip()
            except subprocess.CalledProcessError as e:
                print('Error calling %s: %s' % (j.qsub, e.output.decode().strip()))
            else:
                if not res:
                    print('No qsub output')
                else:
                    j.inputFiles = input_files
                    if 'Your job ' in res:
                        m = re.search(r'Your job (\d*) ', res)
                        assert m
                        res = m.group(1)
                    j.jobId = res
                    j.subTime = time.time()
                    open(scriptRoot + jobid_ext, 'w', encoding="utf-8").write(res)
                    addJobIndex(kwargs.get('batchPath'), j)


def queue_job_details(batchPath=None, running=True, queued=True, warnNotBatch=True):
    """
    Return: list of jobIds, list of job_names, list of list names
    """
    index = loadJobIndex(batchPath)
    if not index:
        print('No existing job index found')
        return []
    if spawn.find_executable("showq") is not None:
        qstat = 'showq'
        running_txt = ' Running '
    else:
        if spawn.find_executable("squeue") is not None:
            qstat = 'squeue'
        else:
            # e.g. Sun Grid Engine/OGS
            qstat = 'qstat'
        running_txt = ' r '
    res_str = subprocess.check_output('%s -u $USER' % qstat, shell=True).decode().strip()
    res = res_str.split("\n")
    names = []
    job_names = []
    ids = []
    infos = []
    for line in res[2:]:
        if ((' ' + str(os.environ.get('USER')) + ' ' in line and
             (queued and not re.search(running_txt, line, re.IGNORECASE) or
              running and re.search(running_txt, line, re.IGNORECASE)))):
            items = line.split()
            jobId = items[0]
            j = index.jobSettings.get(jobId)
            if j is None:
                jobId_items = items[0].split('.')
                if jobId_items[0].upper() == 'TOTAL':
                    continue
                if len(jobId_items) == 1 or jobId_items[0].isdigit():
                    jobId = jobId_items[0]
                else:
                    jobId = jobId_items[1]
                j = index.jobSettings.get(jobId)
            if j is None:
                if warnNotBatch:
                    print('...Job ' + jobId + ' not in this batch, skipping')
                continue
            names += [j.names]
            job_names += [j.jobName]
            ids += [jobId]
            infos += [line]
    return ids, job_names, names, infos


def queue_job_names(batchPath=None, running=False, queued=True):
    lists = queue_job_details(batchPath, running, queued)[2]
    names = []
    for nameset in lists:
        names += nameset
    return names


# Functions to manipulate job templates ##################################################

def replace_placeholders(txt, vals):
    """
    Replaces placeholders ``vals`` (dict) in a template ``txt``.
    """
    txt = txt.replace('\r', '').format(**vals)
    return txt


def extract_value(txt, name):
    """
    Extracts value of variable ``name`` defined in a template ``txt``
    as ``##name: value``.
    """
    match = re.search('##' + name + ':(.*)##', txt)
    if match:
        return match.group(1).strip()
    return None


def get_defaulted(key_name, default=None, tp: Any = str, template=None, ext_env=None,
                  **kwargs) -> Any:
    val = kwargs.get(key_name)
    if val is None and template is not None:
        val = extract_value(template, 'DEFAULT_' + key_name)
    if val is None:
        val = os.environ.get(code_prefix + '_' + key_name, None)
    if val is None and ext_env:
        val = os.environ.get('ext_env', None)
    if val is None:
        val = default
    if val is None:
        return None
    return tp(val)



"""
Tests the facility for importing external priors.

It tests all possible input methods: callable and string
(direct evaluation and ``import_module``).

In each case, it tests the correctness of the values generated, and of the updated info.

The test prior is a gaussian half-ring, combined with a gaussian in one of the tests.

For manual testing, and observing/plotting the density, pass `manual=True`
to `body of test`.
"""
# Local
from .common_external import info_string, info_callable, info_mixed, info_import
from .common_external import body_of_test


def test_prior_external_string(tmpdir):
    body_of_test(info_string, "prior", tmpdir)


def test_prior_external_callable(tmpdir):
    body_of_test(info_callable, "prior", tmpdir)


def test_prior_external_mixed(tmpdir):
    body_of_test(info_mixed, "prior", tmpdir)


def test_prior_external_import(tmpdir):
    body_of_test(info_import, "prior", tmpdir)



import pytest
import logging
from cobaya.model import get_model
from cobaya.theory import Theory
from cobaya.likelihood import Likelihood
from cobaya.log import LoggedError, NoLogging
from cobaya.typing import InputDict

debug = False


# Aderived = 1
# Aout = [Ain]
# Bpar = 3
# Bout = (3, [Ain])
# Bderived = 10
# Cout = Bout

class A(Theory):
    def get_requirements(self):
        return {'Ain'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['Aout'] = [self.provider.get_param('Ain')]
        if want_derived:
            state['derived'] = {'Aderived': 1}

    def get_Aresult(self):
        return self.current_state['Aout']

    def get_can_provide_params(self):
        return ['Aderived']


class B(Theory):
    params = {'Bpar': None, 'Bderived': {'derived': True}}

    def get_requirements(self):
        return {'Aderived', 'Aresult'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['Bout'] = (self.provider.get_param('Aderived') * params_values_dict['Bpar']
                         , self.provider.get_Aresult())

        if want_derived:
            state['derived'] = {'Bderived': 10}

    def get_Bout(self):
        return self.current_state['Bout']


class Balt(Theory):
    params = {'Bpar': None, 'Aderived': None}

    def get_requirements(self):
        return {'Aresult'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['Bout'] = (params_values_dict['Aderived'] * params_values_dict['Bpar']
                         , self.provider.get_Aresult())

    def get_Bout(self):
        return self.current_state['Bout']


class B2(Theory):

    def get_requirements(self):
        return {'Aderived', 'Aresult', 'Bpar'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['Bout'] = (self.provider.get_param('Aderived') * params_values_dict['Bpar'],
                         self.provider.get_Aresult())

        if want_derived:
            state['derived'] = {'Bderived': 10}

    def get_Bout(self):
        return self.current_state['Bout']


class A2(Theory):  # circular
    def get_requirements(self):
        return ('Ain', None), ('Bout', None)

    def get_can_provide_params(self):
        return ['Aderived', 'Aresult']


class C(Theory):  # ambiguous

    def get_requirements(self):
        return {'Aresult'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['Cout'] = (3, [5])

    def get_Bout(self):
        return self.current_state['Cout']


class Like(Likelihood):

    def get_requirements(self):
        return {'Bout'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        res = self.provider.get_Bout()
        state["logp"] = res[0] + res[1][0]


info: InputDict = {'likelihood': {'like': Like},
                   'params': {'Bpar': 3, 'Ain': 5},
                   'debug': debug}


def _test_loglike(theories):
    for th in theories, theories[::-1]:
        info['theory'] = dict(th)
        model = get_model(info)

        assert model.loglikes({})[0] == 8, "test loglike failed for %s" % th
        assert model.loglikes({}, return_derived=False,
                              cached=False) == 8, "non-derived loglike failed for %s" % th


def test_dependencies():
    theories = [('A', A), ('B', B)]
    _test_loglike(theories)
    _test_loglike([('A', A), ('B', B2)])
    _test_loglike([('A', A), ('B', Balt)])

    info['params']['Aderived'] = None
    _test_loglike([('A', A), ('B', Balt)])
    info['params']['Aderived'] = {'derived': True}
    _test_loglike([('A', A), ('B', Balt)])
    del info['params']['Aderived']

    info['params']['Bderived'] = {'derived': True}
    info['theory'] = dict(theories)
    model = get_model(info)
    assert model.loglikes({})[1] == [10], "failed"
    info['params'].pop('Bderived')

    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        _test_loglike([('A', A2), ('B', B)])
    assert "Circular dependency" in str(e.value)

    _test_loglike([('A', {'external': A}), ('B', B2)])

    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        _test_loglike([('A', A), ('B', B2), ('C', C)])
    assert "Bout is provided by more than one component" in str(e.value)

    _test_loglike([('A', A), ('B', B2), ('C', {'external': C, 'provides': 'Bout'})])
    _test_loglike([('A', A), ('B', {'external': B2, 'provides': ['Bout']}),
                   ('C', {'external': C})])

    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        _test_loglike([('A', A), ('B', {'external': B2, 'provides': ['Bout']}),
                       ('C', {'external': C, 'provides': ['Bout']})])
    assert "more than one component provides Bout" in str(e.value)

    inf: InputDict = info.copy()
    inf['params'] = info['params'].copy()
    inf['params']['notused'] = [1, 10, 2, 5, 1]
    inf['theory'] = dict(theories)
    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        get_model(inf)
    assert "Could not find anything to use input parameter" in str(e.value)
    inf['params']['notused'] = [1, 10, 2]
    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        get_model(inf)
    assert "Parameter info length not valid" in str(e.value)


# test conditional requirements
class D(Theory):

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['D'] = self.provider.get_Aresult()[0] * 2

    def get_result(self, result_name, **kwargs):
        if result_name == 'Dresult':
            return self.current_state['D']

    def get_can_provide(self):
        return ['Dresult']

    def must_provide(self, **must_provide):
        if 'Dresult' in must_provide:
            return {'Aresult'}


class E(Theory):

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['E'] = self.provider.get_result('Dresult') * 2

    def get_Eresult(self):
        return self.current_state['E']

    def must_provide(self, **must_provide):
        if 'Eresult' in must_provide:
            return {'Dresult'}


class Like2(Likelihood):

    def get_requirements(self):
        return {'Dresult'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state["logp"] = self.provider.get_result('Dresult') * 2


class Like3(Likelihood):

    def get_requirements(self):
        return {'Eresult'}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state["logp"] = self.provider.get_Eresult()


# circular
class F(Theory):

    def get_Fresult(self):
        pass

    def must_provide(self, **must_provide):
        if 'Fresult' in must_provide:
            return {'LikeDerived'}


class Like4(Likelihood):

    def get_requirements(self):
        return {'Fresult'}

    def get_LikeDerived(self):
        pass


info2: InputDict = {'likelihood': {'like': Like2},
                    'params': {'Ain': 5},
                    'debug': debug, 'stop_at_error': True}


def _test_loglike2(theories):
    for th in theories, theories[::-1]:
        info2['theory'] = dict(th)
        model = get_model(info2)
        assert model.loglike()[0] == 20., "fail conditional dependency for %s" % th


def test_conditional_dependencies():
    theories = [('A', A), ('D', D)]
    _test_loglike2(theories)

    theories = [('A', A), ('D', D), ('E', E)]
    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        _test_loglike2(theories)
    assert "seems not to depend on any parameters" in str(e.value)

    info2['likelihood']['like'] = Like3
    theories = [('A', A), ('D', D), ('E', E)]
    _test_loglike2(theories)

    info2['likelihood']['like'] = Like4
    theories = [('A', A), ('E', E), ('F', F), ('D', D)]
    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        _test_loglike2(theories)
    assert "Circular dependency" in str(e.value)



"""
.. module:: samplers.evaluate

:Synopsis: Dummy "sampler": simply evaluates the likelihood.
:Author: Jesus Torrado

"""

# Global
import numpy as np
from typing import Mapping

# Local
from cobaya.sampler import Sampler
from cobaya.collection import SampleCollection
from cobaya.log import LoggedError


class Evaluate(Sampler):
    file_base_name = 'evaluate'

    override: Mapping[str, float]
    N: int

    def initialize(self):
        """
        Creates a 1-point collection to store the point
        at which the posterior is evaluated.
        """
        try:
            self.N = int(self.N)
        except ValueError:
            raise LoggedError(
                self.log,
                "Could not convert the number of samples to an integer: %r", self.N)
        self.one_point = SampleCollection(self.model, self.output, name="1")
        self.log.info("Initialized!")

    def run(self):
        """
        First gets a reference point. If a single reference point is not given,
        the point is sampled from the reference pdf. If that one is not defined either,
        the point is sampled from the prior.

        Then it evaluates the prior and likelihood(s) and stores them in the one-member
        sample collection.
        """
        for i in range(self.N):
            if self.N > 1:
                self.log.info("Evaluating sample #%d ------------------------------",
                              i + 1)
            self.log.info("Looking for a reference point with non-zero prior.")
            reference_values = self.model.prior.reference(random_state=self._rng)
            reference_point = dict(
                zip(self.model.parameterization.sampled_params(), reference_values))
            for p, v in (self.override or {}).items():
                if p not in reference_point:
                    raise LoggedError(
                        self.log, "Parameter '%s' used in override not known. "
                                  "Known parameters names are %r.",
                        p, self.model.parameterization.sampled_params())
                reference_point[p] = v
            self.log.info("Reference point:\n   " + "\n   ".join(
                ["%s = %g" % pv for pv in reference_point.items()]))
            self.log.info("Evaluating prior and likelihoods...")
            self.logposterior = self.model.logposterior(reference_point)
            self.one_point.add(
                list(reference_point.values()), derived=self.logposterior.derived,
                logpost=self.logposterior.logpost, logpriors=self.logposterior.logpriors,
                loglikes=self.logposterior.loglikes)
            self.log.info("log-posterior  = %g", self.logposterior.logpost)
            self.log.info("log-prior      = %g", self.logposterior.logprior)
            for j, name in enumerate(self.model.prior):
                self.log.info(
                    "   logprior_" + name + " = %g", self.logposterior.logpriors[j])
            if self.logposterior.logprior > -np.inf:
                self.log.info("log-likelihood = %g", self.logposterior.loglike)
                for j, name in enumerate(self.model.likelihood):
                    self.log.info(
                        "   chi2_" + name + " = %g", (-2 * self.logposterior.loglikes[j]))
                self.log.info("Derived params:")
                for name, value in zip(self.model.parameterization.derived_params(),
                                       self.logposterior.derived):
                    self.log.info("   " + name + " = %g", value)
            else:
                self.log.info("Likelihoods and derived parameters not computed, "
                              "since the prior is null.")
        # Write the output: the point and its prior, posterior and likelihood.
        self.one_point.out_update()

    def products(self):
        """
        Auxiliary function to define what should be returned in a scripted call.

        Returns:
           The sample ``SampleCollection`` containing the
           sequentially discarded live points.
        """
        return {"sample": self.one_point}



import os
from copy import deepcopy
from scipy.stats import multivariate_normal
from getdist.mcsamples import loadMCSamples, MCSamplesFromCobaya
import numpy as np
import pytest

from cobaya.run import run
from cobaya.post import post, OutputOptions
from cobaya.typing import ParamsDict, InputDict
from cobaya.conventions import separator_files
from cobaya import mpi

pytestmark = pytest.mark.mpi

_post_ = separator_files + "post" + separator_files

mean = np.array([0, 0])
sigma = 0.5
cov = np.array([[sigma ** 2, 0], [0, sigma ** 2]])
sampled = {"mean": mean, "cov": cov}
target = {"mean": mean + np.array([sigma / 2, 0]), "cov": cov}


def allclose(a, b):
    return np.allclose(a.to_numpy(dtype=np.float64), b.to_numpy(dtype=np.float64))


def sampled_pdf(a, b):
    return multivariate_normal.logpdf([a, b], mean=sampled["mean"], cov=sampled["cov"])


def target_pdf(a, b, c=0):
    logp = multivariate_normal.logpdf([a, b], mean=target["mean"], cov=target["cov"])
    derived = {"cprime": c}
    return logp, derived


def target_pdf_prior(a, b):
    return target_pdf(a, b, c=0)[0]


_range = {"min": -2, "max": 2}
ref_pdf = {"dist": "norm", "loc": 0, "scale": 0.1}
info_params: ParamsDict = dict([
    ("a", {"prior": _range, "ref": ref_pdf, "proposal": sigma}),
    ("b", {"prior": _range, "ref": ref_pdf, "proposal": sigma}),
    ("a_plus_b", {"derived": lambda a, b: a + b})])

info_sampler = {"mcmc": {"Rminus1_stop": 0.25, "Rminus1_cl_stop": 0.5, "seed": 1}}
info_sampler_dummy = {"evaluate": {"N": 10}}


def _get_targets(mcsamples_in):
    loglikes = []
    for a, b in zip(mcsamples_in['a'], mcsamples_in['b']):
        loglikes.append(-target_pdf_prior(a, b) + sampled_pdf(a, b))
    mcsamples_in.reweightAddingLogLikes(loglikes)
    target_mean = mcsamples_in.mean(["a", "b"])
    target_cov = mcsamples_in.getCovMat().matrix
    return target_mean, target_cov


@mpi.sync_errors
@pytest.mark.parametrize("temperature", (1, 2))
def test_post_prior(tmpdir, temperature):
    """
    Swaps prior "gaussian" for "target".

    It also tests loading vs passing samples to post, and different MCMC temperatures.
    """
    # Generate original chain
    info: InputDict = {
        "output": os.path.join(tmpdir, "gaussian"), "force": True,
        "params": info_params, "sampler": deepcopy(info_sampler),
        "likelihood": {"one": None}, "prior": {"gaussian": sampled_pdf}}
    info_post: InputDict = {
        "output": info["output"], "force": True,
        "post": {"suffix": "foo", 'skip': 0.1,
                 "remove": {"prior": {"gaussian": None}},
                 "add": {"prior": {"target": target_pdf_prior}}}}
    if list(info["sampler"].keys())[0] == "mcmc":
        if info["sampler"]["mcmc"] is None:
            info["sampler"]["mcmc"] = {}
        info["sampler"]["mcmc"]["temperature"] = temperature
    _, sampler = run(info)
    if mpi.is_main_process():
        mcsamples_in = loadMCSamples(info["output"], settings={'ignore_rows': 0.1})
        mcsamples_in.cool(temperature)
        target_mean, target_cov = mpi.share(_get_targets(mcsamples_in))
    else:
        target_mean, target_cov = mpi.share()
    for pass_chains in [False, True]:
        post(info_post, sample=sampler.products()["sample"] if pass_chains else None)
        # Load with GetDist and compare
        if mpi.is_main_process():
            mcsamples = loadMCSamples(
                info_post["output"] + _post_ + info_post["post"]["suffix"])
            new_mean = mcsamples.mean(["a", "b"])
            new_cov = mcsamples.getCovMat().matrix
            mpi.share((new_mean, new_cov))
        else:
            new_mean, new_cov = mpi.share()
        # Noisier with higher temperature
        atol, rtol = (1e-8, 1e-5) if temperature == 1 else (5e-4, 1e-3)
        assert np.allclose(new_mean, target_mean, atol=atol, rtol=rtol)
        assert np.allclose(new_cov, target_cov, atol=atol, rtol=rtol)


def test_post_likelihood():
    """
    Swaps likelihood "gaussian" for "target".

    It also tests aggregated chi2's by removing and adding a likelihood to an existing
    type.
    """
    # Generate original chain
    orig_interval = OutputOptions.output_inteveral_s
    try:
        OutputOptions.output_inteveral_s = 0
        info_params_local = deepcopy(info_params)
        info_params_local["dummy"] = 0
        dummy_loglike_add = 0.1
        dummy_loglike_remove = 0.01
        info = {
            "output": None, "force": True,
            "params": info_params_local, "sampler": info_sampler,
            "likelihood": {
                "gaussian": {"external": sampled_pdf, "type": "A"},
                "dummy": {"external": lambda dummy: 1, "type": "BB"},
                "dummy_remove": {"external": lambda dummy: dummy_loglike_add,
                                 "type": "BB"}}}
        info_out, sampler = run(info)
        samples_in = mpi.gather(sampler.products()["sample"])
        if mpi.is_main_process():
            mcsamples_in = MCSamplesFromCobaya(info_out, samples_in)
        else:
            mcsamples_in = None

        info_out.update({
            "post": {"suffix": "foo",
                     "remove": {"likelihood": {
                         "gaussian": None, "dummy_remove": None}},
                     "add": {"likelihood": {
                         "target": {
                             "external": target_pdf, "type": "A",
                             "output_params": ["cprime"]},
                         "dummy_add": {
                             "external": lambda dummy: dummy_loglike_remove,
                             "type": "BB"}}}}})
        info_post_out, products_post = post(info_out, sampler.products()["sample"])
        samples = mpi.gather(products_post["sample"])

        # Load with GetDist and compare
        if mcsamples_in:
            target_mean, target_cov = mpi.share(_get_targets(mcsamples_in))

            mcsamples = MCSamplesFromCobaya(info_post_out, samples, name_tag="sample")
            new_mean = mcsamples.mean(["a", "b"])
            new_cov = mcsamples.getCovMat().matrix
            mpi.share((new_mean, new_cov))
        else:
            target_mean, target_cov = mpi.share()
            new_mean, new_cov = mpi.share()
        assert np.allclose(new_mean, target_mean)
        assert np.allclose(new_cov, target_cov)
        assert allclose(products_post["sample"]["chi2__A"],
                        products_post["sample"]["chi2__target"])
        assert allclose(products_post["sample"]["chi2__BB"],
                        products_post["sample"]["chi2__dummy"] +
                        products_post["sample"]["chi2__dummy_add"])
    finally:
        OutputOptions.output_inteveral_s = orig_interval


def test_post_params():
    # Tests:
    # - added simple dynamical derived parameter "a+b"
    # - added dynamical derived parameter that depends on *new* chi2__target
    # - added new fixed input "c" + new derived-from-external-function "cprime"
    # Generate original chain
    info = {
        "params": info_params, "sampler": info_sampler_dummy,
        "likelihood": {"gaussian": sampled_pdf}}
    updated_info_gaussian, sampler_gaussian = run(info)
    products_gaussian = sampler_gaussian.products()
    info_post = {
        "post": {"suffix": "foo",
                 "remove": {"params": {"a_plus_b": None}},
                 "add": {
                     "likelihood": {
                         "target": {"external": target_pdf, "output_params": ["cprime"]}},
                     "params": {
                         "c": 1.234,
                         "a_minus_b": {"derived": "lambda a,b: a-b"},
                         "my_chi2__target": {
                             "derived": "lambda chi2__target: chi2__target"},
                         "cprime": None}}}}
    info_post.update(updated_info_gaussian)
    _, products = post(info_post, products_gaussian["sample"])
    # Compare parameters
    assert allclose(products["sample"]["a"] - products["sample"]["b"],
                    products["sample"]["a_minus_b"])
    assert np.allclose(products["sample"]["cprime"].to_numpy(dtype=np.float64),
                       info_post["post"]["add"]["params"]["c"])
    assert allclose(products["sample"]["my_chi2__target"],
                    products["sample"]["chi2__target"])



"""
.. module:: output

:Synopsis: Generic output class and output drivers
:Author: Jesus Torrado

"""
# Global
import os
import sys
import datetime
import re
import shutil
from packaging import version
from typing import Optional, Any
# Local
from cobaya.yaml import yaml_dump, yaml_load, yaml_load_file, \
    OutputError, InputImportError
from cobaya.conventions import resume_default, Extension, kinds, get_version
from cobaya.typing import InputDict
from cobaya.log import LoggedError, HasLogger, get_logger, get_traceback_text
from cobaya.input import is_equal_info, load_info_dump, split_prefix, get_info_path
from cobaya.tools import deepcopy_where_possible, find_with_regexp, sort_cosmetic, \
    has_non_yaml_reproducible
from cobaya.component import get_component_class
from cobaya import mpi

# Default output type and extension
_kind = "txt"
_ext = "txt"


def use_portalocker():
    if os.getenv('COBAYA_USE_FILE_LOCKING', 't').lower() in ('true', '1', 't'):
        try:
            import portalocker
        except ModuleNotFoundError:
            return None
        else:
            return True
    return False


class FileLock:
    _file_handle: Any

    def __init__(self, filename=None, log=None):
        self.lock_error_file = ""
        self.lock_file = ""
        if filename:
            self.set_lock(log, filename)

    def set_lock(self, log, filename, force=False):
        if self.has_lock():
            return
        self.lock_file = filename + '.locked'
        self.lock_error_file = filename + '.lock_err'
        try:
            os.remove(self.lock_error_file)
        except OSError:
            pass
        self.log = log or get_logger("file_lock")
        try:
            h: Any = None
            if use_portalocker():
                import portalocker
                try:
                    h = open(self.lock_file, 'wb')
                    portalocker.lock(h, portalocker.LOCK_EX + portalocker.LOCK_NB)
                    self._file_handle = h
                except portalocker.exceptions.BaseLockException:
                    if h:
                        h.close()
                    self.lock_error()
            else:
                # will work, but crashes will leave .lock files that will raise error
                self._file_handle = open(self.lock_file, 'wb' if force else 'xb')
        except OSError:
            self.lock_error()

    def lock_error(self):
        if not self.has_lock():
            assert self.lock_error_file
            try:
                # make lock_err so process holding lock can check
                # another process had an error
                with open(self.lock_error_file, 'wb'):
                    pass
            except OSError:
                pass
        if mpi.is_disabled():
            raise LoggedError(self.log,
                              "File %s is locked by another process, you are running "
                              "with MPI disabled but may have more than one process. "
                              "Note that --test should not be used with MPI.")
        if mpi.get_mpi():
            import mpi4py
        else:
            mpi4py = None
        if mpi.is_main_process() and use_portalocker() is None:
            self.log.warning('install "portalocker" for better file lock control.')
        raise LoggedError(self.log,
                          "File %s is locked.\nYou may be running multiple jobs with "
                          "the same output when you intended to run with MPI. "
                          "Check that mpi4py is correctly installed and "
                          "configured (using the same mpi as mpirun/mpiexec); "
                          "e.g. try the test at\n"
                          "https://cobaya.readthedocs.io/en/latest/installation."
                          "html#mpi-parallelization-optional-but-encouraged\n"
                          + ("Your current mpi4py config is:"
                             "\n %s" % mpi4py.get_config()
                             if mpi4py is not None else
                             "mpi4py is NOT currently installed."), self.lock_file)

    def check_error(self):
        if self.lock_error_file and os.path.exists(self.lock_error_file):
            self.lock_error()

    def clear_lock(self):
        if self.has_lock():
            self._file_handle.close()
            del self._file_handle
            os.remove(self.lock_file)
            try:
                os.remove(self.lock_error_file)
            except OSError:
                pass
        self.lock_error_file = ""
        self.lock_file = ""

    def has_lock(self):
        return hasattr(self, "_file_handle")

    def __del__(self):
        self.clear_lock()


class Output(HasLogger):
    """
    Basic output driver. It takes care of creating the output files, checking
    compatibility with old runs when resuming, cleaning up when forcing, preparing
    :class:`~collection.SampleCollection` files, etc.
    """

    _old_updated_info: Optional[InputDict]

    @mpi.set_from_root(("force", "folder", "prefix", "kind", "ext",
                        "_resuming", "prefix_regexp_str", "log"))
    def __init__(self, prefix, resume=resume_default, force=False, infix=None):
        self.name = "output"
        self.set_logger(self.name)
        self.lock = FileLock()
        self.folder, self.prefix = split_prefix(prefix)
        self.prefix_regexp_str = re.escape(self.prefix) + (
            r"[\._]" if self.prefix else "")
        self.force = force
        if resume and force and prefix:
            # No resume and force at the same time (if output)
            raise LoggedError(
                self.log,
                "Make 'resume: True' or 'force: True', not both at the same time: "
                "can't simultaneously overwrite a chain and resume from it.")
        if not os.path.exists(self.folder):
            self.log.debug("Creating output folder '%s'", self.folder)
            try:
                os.makedirs(self.folder)
            except OSError:
                self.log.error(get_traceback_text(sys.exc_info()))
                raise LoggedError(
                    self.log, "Could not create folder '%s'. "
                              "See traceback on top of this message.", self.folder)
        self.log.info("Output to be read-from/written-into folder '%s', with prefix '%s'",
                      self.folder, self.prefix)
        # Prepare file names, and check if chain exists
        self.file_input = get_info_path(
            self.folder, self.prefix, infix=infix, kind="input")
        self.file_updated = get_info_path(self.folder, self.prefix, infix=infix)
        self.dump_file_updated = get_info_path(
            self.folder, self.prefix, infix=infix, ext=Extension.dill)
        self._resuming = False
        # Output kind and collection extension
        self.kind = _kind
        self.ext = _ext
        if os.path.isfile(self.file_updated):
            self.log.info(
                "Found existing info files with the requested output prefix: '%s'",
                prefix)
            if self.force:
                self.log.info("Will delete previous products ('force' was requested).")
                self.delete_infos()
                # Sampler products will be deleted at sampler initialisation
            elif resume:
                # Only in this case we can be sure that we are actually resuming
                self._resuming = True
                self.log.info("Let's try to resume/load.")

    def is_prefix_folder(self):
        """
        Returns `True` if the output prefix is a bare folder, e.g. `chains/`.
        """
        return bool(self.prefix)

    def separator_if_needed(self, separator):
        """
        Returns the given separator if there is an actual file name prefix (i.e. the
        output prefix is not a bare folder), or an empty string otherwise.

        Useful to add custom suffixes to output prefixes (may want to use
        `Output.add_suffix` for that).
        """
        return separator if self.is_prefix_folder() else ""

    def sanitize_collection_extension(self, extension):
        """
        Returns the `extension` without the leading dot, if given, or the default one
        `Output.ext` otherwise.
        """
        if extension:
            return extension.lstrip(".")
        return self.ext

    def add_suffix(self, suffix, separator="_"):
        """
        Returns the full output prefix (folder and file name prefix) combined with a
        given suffix, inserting a given separator in between (default: `_`) if needed.
        """
        return os.path.join(self.folder,
                            self.prefix + self.separator_if_needed(separator) + suffix)

    @mpi.root_only
    def create_folder(self, folder):
        """
        Creates the given folder (MPI-aware).
        """
        try:
            if not os.path.exists(folder):
                os.makedirs(folder)
        except Exception as e:
            raise LoggedError(
                self.log, "Could not create folder %r. Reason: %r", folder, str(e))

    @mpi.root_only
    def delete_infos(self):
        self.check_lock()
        for f in [self.file_input, self.file_updated, self.dump_file_updated]:
            try:
                os.remove(f)
            except OSError:
                pass

    def updated_prefix(self):
        """
        Updated path: drops folder: now it's relative to the chain's location.
        """
        return self.prefix or "."

    def is_resuming(self):
        return self._resuming

    @mpi.set_from_root("_resuming")
    def set_resuming(self, value):
        self._resuming = value

    @mpi.from_root
    def load_updated_info(self, cache=False, use_cache=False) -> Optional[InputDict]:
        return self.reload_updated_info(cache=cache, use_cache=use_cache)

    def reload_updated_info(self, cache=False, use_cache=False) -> Optional[InputDict]:
        if mpi.is_main_process():
            if use_cache and hasattr(self, "_old_updated_info"):
                return self._old_updated_info
            try:
                if os.path.isfile(self.dump_file_updated):
                    loaded = load_info_dump(self.dump_file_updated)
                else:
                    loaded = yaml_load_file(self.file_updated)  # type: ignore
                if cache:
                    self._old_updated_info = deepcopy_where_possible(loaded)
                return loaded
            except IOError:
                if cache:
                    self._old_updated_info = None
                return None
        else:
            # Only cached possible when non main process
            if not use_cache:
                raise LoggedError(self.log, "Cannot call `reload_updated_info` from "
                                            "non-main process unless cached version "
                                            "(`use_cache=True`) requested.")
            return getattr(self, "_old_updated_info", None)

    @mpi.set_from_root("_old_updated_info")
    def check_and_dump_info(self, input_info, updated_info, check_compatible=True,
                            cache_old=False, use_cache_old=False, ignore_blocks=()):
        """
        Saves the info in the chain folder twice:
           - the input info.
           - idem, populated with the components' defaults.

        If resuming a sample, checks first that old and new infos and versions are
        consistent.
        """
        # trim known params of each likelihood: for internal use only
        self.check_lock()
        updated_info_trimmed = deepcopy_where_possible(updated_info)
        updated_info_trimmed["version"] = get_version()
        for like_info in updated_info_trimmed.get("likelihood", {}).values():
            (like_info or {}).pop("params", None)
        if check_compatible:
            # We will test the old info against the dumped+loaded new info.
            # This is because we can't actually check if python objects do change
            try:
                old_info = self.reload_updated_info(cache=cache_old,
                                                    use_cache=use_cache_old)
            except InputImportError:
                # for example, when there's a dynamically generated class that cannot
                # be found by the yaml loader (could use yaml loader that ignores them)
                old_info = None
            if old_info:
                # use consistent yaml read-in types
                # TODO: could probably just compare full infos here, with externals?
                #  for the moment cautiously keeping old behaviour
                old_info = yaml_load(yaml_dump(old_info))  # type: ignore
                if old_info.get("test"):
                    old_info = None
            if old_info:
                new_info = yaml_load(yaml_dump(updated_info_trimmed))
                if not is_equal_info(old_info, new_info, strict=False,
                                     ignore_blocks=list(ignore_blocks) + [
                                         "output"]):
                    raise LoggedError(
                        self.log, "Old and new run information not compatible! "
                                  "Resuming not possible!")
                # Deal with version comparison separately:
                # - If not specified now, take the one used in resume info
                # - If specified both now and before, check new older than old one
                # (For Cobaya's own version, prefer new one always)
                old_version = old_info.get("version")
                new_version = new_info.get("version")
                if isinstance(old_version, str) and isinstance(new_version, str):
                    if version.parse(old_version) > version.parse(new_version):
                        raise LoggedError(
                            self.log, "You are trying to resume a run performed with a "
                                      "newer version of Cobaya: %r (you are using %r). "
                                      "Please, update your Cobaya installation.",
                            old_version, new_version)
                for k in set(kinds).intersection(updated_info):
                    if k in ignore_blocks or updated_info[k] is None:
                        continue
                    for c in updated_info[k]:
                        new_version = updated_info[k][c].get("version")
                        old_version = old_info[k][c].get("version")  # type: ignore
                        if new_version is None:
                            updated_info[k][c]["version"] = old_version
                            updated_info_trimmed[k][c]["version"] = old_version
                        elif old_version is not None:
                            cls = get_component_class(
                                c, k, class_name=updated_info[k][c].get("class"),
                                logger=self.log)
                            if cls and cls.compare_versions(
                                    old_version, new_version, equal=False):
                                raise LoggedError(
                                    self.log, "You have requested version %r for "
                                              "%s:%s, but you are trying to resume a "
                                              "run that used a newer version: %r.",
                                    new_version, k, c, old_version)
        # If resuming, we don't want to do *partial* dumps
        if ignore_blocks and self.is_resuming():
            return
        # Work on a copy of the input info, since we are updating the prefix
        # (the updated one is already a copy)
        if input_info is not None:
            input_info = deepcopy_where_possible(input_info)
        # Write the new one
        for f, info in [(self.file_input, input_info),
                        (self.file_updated, updated_info_trimmed)]:
            if info:
                for k in ignore_blocks:
                    info.pop(k, None)
                info.pop("debug", None)
                info.pop("force", None)
                info.pop("resume", None)
                # make sure the dumped output prefix does only contain the file prefix,
                # not the folder, since it's already been placed inside it
                info["output"] = self.updated_prefix()
                with open(f, "w", encoding="utf-8") as f_out:
                    try:
                        f_out.write(yaml_dump(sort_cosmetic(info)))
                    except OutputError as e:
                        raise LoggedError(self.log, str(e))
        if updated_info_trimmed and has_non_yaml_reproducible(updated_info_trimmed):
            try:
                import dill
            except ImportError:
                self.mpi_info('Install "dill" to save reproducible options file.')
            else:
                import pickle
                try:
                    with open(self.dump_file_updated, 'wb') as f:
                        dill.dump(sort_cosmetic(updated_info_trimmed), f,
                                  pickle.HIGHEST_PROTOCOL)
                except pickle.PicklingError as e:
                    os.remove(self.dump_file_updated)
                    self.mpi_info('Options file cannot be pickled %s', e)

    def delete_with_regexp(self, regexp, root=None):
        """
        Deletes all files compatible with the given `regexp`.

        If `regexp` is `None` and `root` is defined, deletes the `root` folder.
        """
        if root is None:
            root = self.folder
        if regexp is not None:
            file_names = find_with_regexp(regexp, root)
            if file_names:
                self.log.debug(
                    "From regexp %r in folder %r, deleting files %r", regexp.pattern,
                    root, file_names)
        else:
            file_names = [root]
            self.log.debug("Deleting folder %r", root)
        for f in file_names:
            self.delete_file_or_folder(f)

    def delete_file_or_folder(self, filename):
        """
        Deletes a file or a folder. Fails silently.
        """
        self.check_lock()
        try:
            os.remove(filename)
        except IsADirectoryError:
            try:
                shutil.rmtree(filename)
            except:
                raise
        except OSError:
            pass

    def prepare_collection(self, name=None, extension=None):
        """
        Generates a file name for the collection, as
        ``[folder]/[prefix].[name].[extension]``.

        Notice that ``name=None`` generates a date, but ``name=""`` removes the ``name``
        field, making it simply ``[folder]/[prefix].[extension]``.
        """
        if name is None:
            name = (datetime.datetime.now().isoformat().replace("T", "")
                        .replace(":", "").replace(".", "")
                        .replace("-", "")[:(4 + 2 + 2) + (2 + 2 + 2 + 3)])  # up to ms
        file_name = os.path.join(
            self.folder,
            self.prefix + ("." if self.prefix else "") + (name + "." if name else "") +
            self.sanitize_collection_extension(extension))
        return file_name, self.kind

    def collection_regexp(self, name=None, extension=None):
        """
        Returns a regexp for collections compatible with this output settings.

        Use `name` for particular types of collections (default: any number).
        Pass `False` to mean there is nothing between the output prefix and the extension.
        """
        if name is None:
            name = r"\d+\."
        elif name is False:
            name = ""
        else:
            name = re.escape(name) + r"\."
        extension = self.sanitize_collection_extension(extension)
        return re.compile(self.prefix_regexp_str + name +
                          re.escape(extension.lower()) + "$")

    def is_collection_file_name(self, file_name, name=None, extension=None):
        """
        Check if a `file_name` is a collection compatible with this `Output` instance.

        Use `name` for particular types of collections (default: any number).
        Pass `False` to mean there is nothing between the output prefix and the extension.
        """
        return (file_name ==
                getattr(self.collection_regexp(name=name, extension=extension)
                        .match(file_name), "group", lambda: None)())

    @mpi.root_only
    def clear_lock(self):
        self.lock.clear_lock()

    @mpi.root_only
    def check_lock(self):
        self.lock.check_error()

    def find_collections(self, name=None, extension=None):
        """
        Returns all collection files found which are compatible with this `Output`
        instance, including their path in their name.

        Use `name` for particular types of collections (default: matches any number).
        Pass `False` to mean there is nothing between the output prefix and the extension.
        """
        return sorted(
            f2 for f2 in [os.path.join(self.folder, f) for f in os.listdir(self.folder)]
            if self.is_collection_file_name(
                os.path.split(f2)[1], name=name, extension=extension))

    def load_collections(self, model, skip=0, thin=1, concatenate=False,
                         name=None, extension=None):
        """
        Loads all collection files found which are compatible with this `Output`
        instance, including their path in their name.

        Use `name` for particular types of collections (default: any number).
        Pass `False` to mean there is nothing between the output prefix and the extension.
        """
        self.check_lock()
        filenames = self.find_collections(name=name, extension=extension)
        # pylint: disable=import-outside-toplevel
        from cobaya.collection import SampleCollection
        collections = [
            SampleCollection(model, self, name="%d" % (1 + i), file_name=filename,
                             load=True, onload_skip=skip, onload_thin=thin)
            for i, filename in enumerate(filenames)]
        if concatenate and collections:
            collection = collections[0]
            for collection_i in collections[1:]:
                collection._append(collection_i)
            return collection
        return collections

    @mpi.root_only
    def set_lock(self):
        self.lock.set_lock(self.log, self.file_input, force=self.force)

    def __enter__(self):
        self.set_lock()
        return self

    def __exit__(self, *args):
        self.clear_lock()


# noinspection PyMissingConstructor
class OutputDummy(Output):
    """
    Dummy output class. Does nothing. Evaluates to 'False' as a class.
    """

    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        self.set_logger()
        self.log.debug("No output requested. Doing nothing.")
        # override all methods that actually produce output
        exclude = ["nullfunc"]
        _func_name = "__name__"
        for attrname, attr in list(Output.__dict__.items()):
            func_name = getattr(attr, _func_name, None)
            if func_name and func_name not in exclude and '__' not in func_name:
                setattr(self, attrname, self.nullfunc)

    def nullfunc(self, *args, **kwargs):
        pass

    def __nonzero__(self):
        return False

    def __bool__(self):
        return False


def get_output(*args, **kwargs) -> Output:
    """
    Auxiliary function to retrieve the output driver
    (e.g. whether to get the MPI-wrapped one, or a dummy output driver).
    """
    if kwargs.get("prefix"):
        return Output(*args, **kwargs)
    else:
        return OutputDummy(*args, **kwargs)



"""
Tests if the initial covariance matrix of the MCMC proposal is build with the right
inheritance order.
"""

import os
import numpy as np
from itertools import chain

from cobaya.likelihoods.gaussian_mixture import random_cov
from cobaya.typing import InputDict
from cobaya.run import run
from cobaya.sampler import CovmatSampler
from cobaya import mpi


def test_mcmc_initial_covmat_interactive():
    dim = 40
    body_of_test(dim)


def test_mcmc_initial_covmat_yaml(tmpdir):
    dim = 40
    body_of_test(dim, tmpdir=tmpdir)


def body_of_test(dim, tmpdir=None, random_state=None):
    mindim = 4
    assert dim > mindim, "Needs dimension>%d for the test." % mindim
    if mpi.is_main_process():
        random_state = np.random.default_rng(random_state)
        i_s = list(range(dim))
        random_state.shuffle(i_s)
        initial_random_covmat = random_cov(dim * [[0, 1]], random_state=random_state)
        mpi.share((i_s, initial_random_covmat))
    else:
        i_s, initial_random_covmat = mpi.share()

    n_altered = int(dim / 4)
    i_proposal = i_s[:n_altered]
    i_ref = i_s[n_altered:2 * n_altered]
    i_prior = i_s[2 * n_altered:3 * n_altered]
    removed = list(chain(*(i_proposal, i_ref, i_prior)))
    i_covmat = [i for i in range(dim) if i not in removed]
    for i in removed:
        diag = initial_random_covmat[i, i]
        initial_random_covmat[:, i] = 0
        initial_random_covmat[i, :] = 0
        initial_random_covmat[i, i] = diag
    # Prepare info, including refs, priors and reduced covmat
    prefix = "a_"
    if mpi.is_main_process():
        input_order = list(range(dim))
        random_state.shuffle(input_order)
    else:
        input_order = None
    input_order = mpi.share(input_order)
    info: InputDict = {"likelihood": {"one": None}, "params": {}}
    fallback_scale = np.sqrt(CovmatSampler.fallback_covmat_scale)
    for i in input_order:
        p = prefix + str(i)
        info["params"][p] = {"prior": {"dist": "norm", "loc": 0, "scale": 1000}}
        sigma = np.sqrt(initial_random_covmat[i, i])
        if i in i_proposal:
            info["params"][p]["proposal"] = sigma
        elif i in i_ref:
            info["params"][prefix + str(i)]["ref"] = {"dist": "norm",
                                                      "scale": sigma * fallback_scale}
        elif i in i_prior:
            info["params"][prefix + str(i)]["prior"]["scale"] = sigma * fallback_scale
    reduced_covmat = initial_random_covmat[np.ix_(i_covmat, i_covmat)]
    reduced_covmat_params = [prefix + str(i) for i in i_covmat]
    info["sampler"] = {"mcmc": {}}
    if tmpdir:
        filename = os.path.join(str(tmpdir), "mycovmat.dat")
        header = " ".join(reduced_covmat_params)
        np.savetxt(filename, reduced_covmat, header=header)
        info["sampler"]["mcmc"]["covmat"] = str(filename)
    else:
        info["sampler"]["mcmc"]["covmat_params"] = reduced_covmat_params
        info["sampler"]["mcmc"]["covmat"] = reduced_covmat
    to_compare = initial_random_covmat[np.ix_(input_order, input_order)]

    def callback(sampler):
        assert np.allclose(to_compare, sampler.proposer.get_covariance())

    info["sampler"]["mcmc"].update({
        "callback_function": callback, "callback_every": 1, "max_samples": 1,
        "burn_in": 0})
    run(info)



"""
.. module:: bicep_keck_2018

:Synopsis: Likelihood of BICEP/Keck, October 2021 (2018 data)
:Author: BICEP/Keck Collaboration and Antony Lewis

.. |br| raw:: html

   <br />

.. note::

   **If you use this likelihood, please cite it as:**
   |br|
   Keck Array and BICEP2 Collaborations,
   `BICEP / Keck XIII: Improved Constraints on Primordial Gravitational Waves
   using Planck, WMAP, and BICEP/Keck Observations through the 2018 Observing Season`
   `(arXiv:2110.00483v1) <https://arxiv.org/abs/2110.00483>`_

Usage
-----

To use this likelihood, ``bicep_keck_2018``, you simply need to mention it in the
``likelihood`` block, or add them using the :doc:`input generator <cosmo_basic_runs>`.

The corresponding nuisance parameters will be added automatically,
so you don't have to care about listing them in the ``params`` block.

The nuisance parameters and their default priors are reproduced below, in the *defaults*
``yaml`` file.

You shouldn't need to modify any of the options of this simple likelihood,
but if you really need to, just copy the ``likelihood`` block into your input ``yaml``
file and modify whatever options you want (you can delete the rest).

.. literalinclude:: ../cobaya/likelihoods/bicep_keck_2018/bicep_keck_2018.yaml
   :language: yaml


Installation
------------

This likelihood can be installed automatically as explained in :doc:`installation_cosmo`.
If are following the instructions there (you should!), you don't need to read the rest
of this section.

Manual installation of the BICEP/Keck likelihood
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Assuming you are installing all your
likelihoods under ``/path/to/likelihoods``, simply do

.. code:: bash

   $ cd /path/to/likelihoods
   $ mkdir bicep_keck_2018
   $ cd bicep_keck_2018
   $ wget http://bicepkeck.org/BK18_datarelease/BK18_cosmomc.tgz
   $ tar xvf BK18_cosmomc.tgz
   $ rm BK18_cosmomc.tgz

After this, mention the path to this likelihood when you include it in an input file as

.. code-block:: yaml

   likelihood:
     bicep_keck_2018:
       path: /path/to/likelihoods/bicep_keck_2018

"""

import numpy as np
from typing import Any

# Local
from cobaya.likelihoods.base_classes import CMBlikes
from cobaya.conventions import Const

# Physical constants
Ghz_Kelvin = Const.h_J_s / Const.kB_J_K * 1e9
T_CMB_K = 2.7255  # fiducial CMB temperature


class bicep_keck_2018(CMBlikes):
    r"""
    CMB power spectrum likelihood of BICEP/Keck XIII, \cite{BK18}.
    """
    install_options = {
        "download_url": r"http://bicepkeck.org/BK18_datarelease/BK18_cosmomc.tgz"}

    lform_dust_decorr: str
    lform_sync_decorr: str

    def init_params(self, ini):
        super().init_params(ini)
        self.fpivot_dust = ini.float('fpivot_dust', 353.0)
        self.fpivot_sync = ini.float('fpivot_sync', 23.0)
        self.bandpasses = []
        for i, used in enumerate(self.used_map_order):
            self.bandpasses += [
                self.read_bandpass(ini.relativeFileName('bandpass[%s]' % used))]

        self.fpivot_dust_decorr = [ini.array_float('fpivot_dust_decorr', i, default) for
                                   i, default in zip([1, 2], [217., 353.])]
        self.fpivot_sync_decorr = [ini.array_float('fpivot_sync_decorr', i, default) for
                                   i, default in zip([1, 2], [22., 33.])]

    def read_bandpass(self, fname):
        bandpass: Any = Bandpass()
        bandpass.R = np.loadtxt(fname)
        nu = bandpass.R[:, 0]
        bandpass.dnu = np.hstack(
            ((nu[1] - nu[0]), (nu[2:] - nu[:-2]) / 2, (nu[-1] - nu[-2])))
        # Calculate thermodynamic temperature conversion between this bandpass
        # and pivot frequencies 353 GHz (used for dust) and 150 GHz (used for sync).
        th_int = np.sum(bandpass.dnu * bandpass.R[:, 1] * bandpass.R[:, 0] ** 4 *
                        np.exp(Ghz_Kelvin * bandpass.R[:, 0] / T_CMB_K) /
                        (np.exp(Ghz_Kelvin * bandpass.R[:, 0] / T_CMB_K) - 1) ** 2)
        nu0 = self.fpivot_dust
        th0 = (nu0 ** 4 * np.exp(Ghz_Kelvin * nu0 / T_CMB_K) /
               (np.exp(Ghz_Kelvin * nu0 / T_CMB_K) - 1) ** 2)
        bandpass.th_dust = th_int / th0
        nu0 = self.fpivot_sync
        th0 = (nu0 ** 4 * np.exp(Ghz_Kelvin * nu0 / T_CMB_K) /
               (np.exp(Ghz_Kelvin * nu0 / T_CMB_K) - 1) ** 2)
        bandpass.th_sync = th_int / th0
        # Calculate bandpass center-of-mass (i.e. mean frequency).
        bandpass.nu_bar = np.dot(bandpass.dnu,
                                 bandpass.R[:, 0] * bandpass.R[:, 1]) / np.dot(
            bandpass.dnu,
            bandpass.R[:, 1])

        return bandpass

    def dust_scaling(self, beta, Tdust, bandpass, nu0, bandcenter_err):
        """Calculates greybody scaling of dust signal defined at 353 GHz
        to specified bandpass."""
        gb_int = np.sum(bandpass.dnu * bandpass.R[:, 1] * bandpass.R[:, 0] ** (3 + beta) /
                        (np.exp(Ghz_Kelvin * bandpass.R[:, 0] / Tdust) - 1))
        # Calculate values at pivot frequency.
        gb0 = nu0 ** (3 + beta) / (np.exp(Ghz_Kelvin * nu0 / Tdust) - 1)
        #  Add correction for band center error
        if bandcenter_err != 1:
            nu_bar = Ghz_Kelvin * bandpass.nu_bar
            # Conversion factor error due to bandcenter error.
            th_err = bandcenter_err ** 4 * (np.exp(Ghz_Kelvin * bandpass % nu_bar *
                                                   (bandcenter_err - 1) / T_CMB_K) *
                                            (np.exp(nu_bar / T_CMB_K) - 1) ** 2 /
                                            (np.exp(nu_bar * bandcenter_err
                                                    / T_CMB_K) - 1) ** 2)
            # Greybody scaling error due to bandcenter error.
            gb_err = bandcenter_err ** (3 + beta) * (np.exp(nu_bar / Tdust) - 1) / \
                     (np.exp(nu_bar * bandcenter_err / Tdust) - 1)
        else:
            th_err = 1
            gb_err = 1

        # Calculate dust scaling.
        return (gb_int / gb0) / bandpass.th_dust * (gb_err / th_err)

    def sync_scaling(self, beta, bandpass, nu0, bandcenter_err):
        """Calculates power-law scaling of synchrotron signal defined at 150 GHz
        to specified bandpass."""
        # Integrate power-law scaling and thermodynamic temperature conversion
        # across experimental bandpass.
        pl_int = np.sum(bandpass.dnu * bandpass.R[:, 1] * bandpass.R[:, 0] ** (2 + beta))
        # Calculate values at pivot frequency.
        pl0 = nu0 ** (2 + beta)
        if bandcenter_err != 1:
            nu_bar = Ghz_Kelvin * bandpass.nu_bar
            th_err = bandcenter_err ** 4 * \
                     np.exp(nu_bar * (bandcenter_err - 1) / T_CMB_K) * \
                     (np.exp(nu_bar / T_CMB_K) - 1) ** 2 / \
                     (np.exp(nu_bar * bandcenter_err / T_CMB_K) - 1) ** 2
            pl_err = bandcenter_err ** (2 + beta)
        else:
            pl_err = 1
            th_err = 1
        # Calculate sync scaling.
        return (pl_int / pl0) / bandpass.th_sync * (pl_err / th_err)

    def decorrelation(self, delta, nu0, nu1, nupivot, rat, lform):
        # Calculate factor by which foreground (dust or sync) power is decreased
        # for a cross-spectrum between two different frequencies.

        # Decorrelation scales as log^2(nu0/nu1). rat is l/l_pivot
        scl_nu = (np.log(nu0 / nu1) ** 2) / (np.log(nupivot[0] / nupivot[1]) ** 2)
        # Functional form for ell scaling is specified in .dataset file.
        if lform == "flat":
            scl_ell = 1.0
        elif lform == 'lin':
            scl_ell = rat
        elif lform == 'quad':
            scl_ell = rat ** 2
        else:
            scl_ell = 1.0

        # Even for small cval, correlation can become negative for sufficiently
        # large frequency difference or ell value (with linear or quadratic scaling).
        # Following Vansyngel et al, A&A, 603, A62 (2017), we use an exponential function
        # to remap the correlation coefficient on to the range [0,1].
        # We symmetrically extend this function to (non-physical) correlation coefficients
        # greater than 1 -- this is only used for validation tests of the likelihood
        # model.
        if delta > 1:
            return 2.0 - np.exp(np.log(2.0 - delta) * scl_nu * scl_ell)
        else:
            return np.exp(np.log(delta) * scl_nu * scl_ell)

    def add_foregrounds(self, cls, data_params):
        lpivot = 80.0
        Adust = data_params['BBdust']
        Async = data_params['BBsync']
        alphadust = data_params['BBalphadust']
        betadust = data_params['BBbetadust']
        Tdust = data_params['BBTdust']
        alphasync = data_params['BBalphasync']
        betasync = data_params['BBbetasync']
        dustsync_corr = data_params['BBdustsynccorr']
        EEtoBB_dust = data_params['EEtoBB_dust']
        EEtoBB_sync = data_params['EEtoBB_sync']
        delta_dust = data_params['delta_dust']
        delta_sync = data_params['delta_sync']
        gamma_corr = data_params['gamma_corr']  # 13

        # Calculate dust and sync scaling for each map.
        bandcenter_err = np.empty(self.nmaps_required)
        fdust = np.empty(self.nmaps_required)
        fsync = np.empty(self.nmaps_required)
        for i, mapname in enumerate(self.used_map_order):
            # Read and assign values to band center error params
            if '95' in mapname:
                bandcenter_err[i] = gamma_corr + data_params['gamma_95'] + 1
            elif '150' in mapname:
                bandcenter_err[i] = gamma_corr + data_params['gamma_150'] + 1
            elif '220' in mapname:
                bandcenter_err[i] = gamma_corr + data_params['gamma_220'] + 1
            else:
                bandcenter_err[i] = 1
            fdust[i] = self.dust_scaling(
                betadust, Tdust, self.bandpasses[i], self.fpivot_dust, bandcenter_err[i])
            fsync[i] = self.sync_scaling(
                betasync, self.bandpasses[i], self.fpivot_sync, bandcenter_err[i])

        rat = np.arange(self.pcl_lmin, self.pcl_lmax + 1) / lpivot
        dustpow = Adust * rat ** alphadust
        syncpow = Async * rat ** alphasync
        dustsyncpow = (dustsync_corr * np.sqrt(Adust * Async) *
                       rat ** ((alphadust + alphasync) / 2))

        #  Only calculate foreground decorrelation if necessary.
        need_sync_decorr = np.abs(delta_sync - 1) > 1e-5
        need_dust_decorr = np.abs(delta_dust - 1) > 1e-5
        for i in range(self.nmaps_required):
            for j in range(i + 1):
                CL = cls[i, j]
                EE = CL.theory_ij[0] == 1 and CL.theory_ij[1] == 1
                BB = CL.theory_ij[0] == 2 and CL.theory_ij[1] == 2

                if EE or BB:
                    dust = fdust[i] * fdust[j]
                    sync = fsync[i] * fsync[j]
                    dustsync = fdust[i] * fsync[j] + fsync[i] * fdust[j]

                    if EE:
                        # EE spectrum: multiply foregrounds by EE/BB ratio
                        dust *= EEtoBB_dust
                        sync *= EEtoBB_sync
                        dustsync *= np.sqrt(EEtoBB_sync * EEtoBB_dust)

                    if need_dust_decorr and i != j:
                        corr_dust = self.decorrelation(delta_dust,
                                                       self.bandpasses[i].nu_bar *
                                                       bandcenter_err[i],
                                                       self.bandpasses[j].nu_bar *
                                                       bandcenter_err[j],
                                                       self.fpivot_dust_decorr, rat,
                                                       self.lform_dust_decorr)
                    else:
                        corr_dust = 1
                    if need_sync_decorr and i != j:
                        corr_sync = self.decorrelation(delta_sync,
                                                       self.bandpasses[i].nu_bar *
                                                       bandcenter_err[i],
                                                       self.bandpasses[j].nu_bar *
                                                       bandcenter_err[j],
                                                       self.fpivot_sync_decorr, rat,
                                                       self.lform_sync_decorr)
                    else:
                        corr_sync = 1
                    #  Add foreground model to theory spectrum.
                    # NOTE: Decorrelation is not implemented for the dust/sync
                    # correlated component.
                    # In BK15, we never turned on correlation and decorrelation parameters
                    # simultaneously.
                    CL.CL += dust * dustpow * corr_dust + sync * syncpow * corr_sync \
                             + dustsync * dustsyncpow


class Bandpass:
    pass



import numpy as np
from cobaya.likelihoods.base_classes.planck_2018_CamSpec_python import \
    Planck2018CamSpecPython


class Planck2018CamSpec2021Python(Planck2018CamSpecPython):
    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1", "asset": "CamSpec2021.zip",
                       "directory": "planck_2018_CamSpec2021"}

    bibtex_file = 'CamSpec2021.bibtex'

    def get_powerlaw_residuals(self, data_params):

        amp = np.empty(4)
        amp[0] = data_params['amp_100']
        amp[1] = data_params['amp_143']
        amp[2] = data_params['amp_217']
        amp[3] = data_params['amp_143x217']

        tilt = np.empty(4)
        tilt[0] = data_params['n_100']
        tilt[1] = data_params['n_143']
        tilt[2] = data_params['n_217']
        tilt[3] = data_params['n_143x217']

        powerlaw_pivot = 1500
        C_powerlaw = np.array(
            [amp[ii] * (self.ls / powerlaw_pivot) ** tilt[ii] for ii in range(4)])

        return C_powerlaw

    def get_foregrounds(self, data_params):
        use_fg_residual_model = data_params['use_fg_residual_model']
        if use_fg_residual_model == 0:
            foregrounds = self.get_powerlaw_residuals(data_params)
        elif use_fg_residual_model == 1:
            foregrounds = super().get_foregrounds(data_params)
        elif use_fg_residual_model == 2:
            foregrounds = super().get_foregrounds(
                data_params) + self.get_powerlaw_residuals(data_params)
        else:
            raise ValueError("use_fg_residual_model should be 0 (powerlaw), "
                             "1 (foregrounds) or 2 (both)")
        return foregrounds
